from typing import List, Tuple, Optional
import numpy as np
import scanpy as sc
import pandas as pd
import os

def construct_gem_expressions(adata: sc.AnnData,
                            gem_expr_key: str = 'X_gem',
                            scale_gem_expr: bool = True,
                            layer_key: Optional[str] = None):
    
    gem_expressions = adata.obsm[gem_expr_key]

    # Scale so that the GEM memberships sum to 1 per cell
    gem_sum = gem_expressions.sum(axis=0)
    gem_expressions = gem_expressions / gem_sum
    
    num_gems = gem_expressions.shape[1]
    flow_gems = ['GEM-' + str(i + 1) for i in range(num_gems)]

    adata_gem = sc.AnnData(X=gem_expressions)
    adata_gem.var.index = pd.Index(flow_gems)
    adata_gem.var['downstream_tfs'] = '' # For housekeeping for later
    adata_gem.var['type'] = 'module' # Define variable types
    adata_gem.var['interactions'] = '' # For housekeeping for later

    if scale_gem_expr:

        if layer_key is not None:

            scale_factor = adata.layers[layer_key].copy().sum(1).mean()

        else:
            scale_factor = np.expm1(adata.X).sum(1).mean()

        adata_gem.X *= scale_factor
        sc.pp.log1p(adata_gem)

    return adata_gem, flow_gems

def construct_inflow_signals_cellchat(adata: sc.AnnData,
                                    cellchat_output_key: str, 
                                    model_organism: str = 'human'):
    
    model_organisms = ['human', 'mouse']

    if model_organism not in model_organisms:
        raise ValueError ("Invalid model organism. Please select one of: %s" % model_organisms)
    
    data_dir = os.path.join(os.path.dirname(__file__), '..', 'data')
    data_path = os.path.join(data_dir, 'cellchat_interactions_and_tfs_' + model_organism + '.csv')

    cellchat_interactions_and_tfs = pd.read_csv(data_path, index_col=0)

    ccc_output_merged = pd.concat([adata.uns[cellchat_output_key][sample] for sample in adata.uns[cellchat_output_key]])
    ccc_interactions = ccc_output_merged['interaction_name_2'].unique().tolist()

    unique_inflow_vars_and_interactions = {}

    # The inflow variables are constructed from receptor gene expression
    # that will be weighted by the average expression of their downstream
    # TF targets. These can be multi-units, but because of the weighting,
    # we hestitate to call them receptors in the context of cellular flows.
    for i, interaction in enumerate(ccc_interactions):
        receptor = interaction.split(' - ')[1].strip('()')
        receptor_split = receptor.split('+')
        receptors = []
        
        for i, rec in enumerate(receptor_split):
            if rec not in adata.var_names:
                receptor_v2_split = ccc_output_merged[ccc_output_merged['interaction_name_2'] == interaction]['receptor'].unique()[0].split('_')
                
                receptors.append(receptor_v2_split[i])
            else:
                receptors.append(rec)
        
        receptor = '+'.join(receptors)

        if receptor not in unique_inflow_vars_and_interactions:
            unique_inflow_vars_and_interactions[receptor] = [interaction]
            
        else:
            interactions_for_receptor = unique_inflow_vars_and_interactions[receptor]
            interactions_for_receptor.append(interaction)
            unique_inflow_vars_and_interactions[receptor] = interactions_for_receptor
            
            
    inflow_vars = sorted(list(unique_inflow_vars_and_interactions.keys()))
    num_inflow_vars = len(inflow_vars)

    inflow_expressions = np.zeros((adata.n_obs, num_inflow_vars)) 

    for i, receptor in enumerate(inflow_vars):
        
        receptor_expression = np.ones((adata.n_obs, ))
        
        split_receptor = receptor.split('+')
        considered_receptors = []
        
        for unit in split_receptor:
            
            if unit not in adata.var_names:
                print(unit)
                
            else:
                
                unit_expression = adata[:, unit].X.toarray().flatten()

                receptor_expression *= unit_expression
                considered_receptors.append(unit)
                
        if len(considered_receptors) != 0:
            inflow_expressions[:, i] = receptor_expression**(1.0 / len(considered_receptors))
        
    inflow_expressions_adjusted = inflow_expressions.copy()

    inflow_interactions = []
    unique_inflow_vars_and_tfs = {}

    for i, receptor in enumerate(inflow_vars):
        
        interactions_for_receptor = unique_inflow_vars_and_interactions[receptor]
        
        relevant_interactions = cellchat_interactions_and_tfs[cellchat_interactions_and_tfs['interaction_name_2'].isin(interactions_for_receptor)]
        
        joined_interactions = '/'.join(sorted(interactions_for_receptor))
        inflow_interactions.append(joined_interactions)
        
        downstream_tfs = []
        
        # Go through each category
        possible_downstream_tfs = relevant_interactions['Receptor-TF-combined'].dropna().tolist()\
                                    +  relevant_interactions['Ligand-TF-combined'].dropna().tolist()
        
        for unit in possible_downstream_tfs:
            split_unit = unit.split('_')
            for tf in split_unit:
                if tf not in downstream_tfs:
                    downstream_tfs.append(tf)
                    
        downstream_tfs = np.intersect1d(downstream_tfs, adata.var_names)
        
        unique_inflow_vars_and_tfs[receptor] = sorted(list(downstream_tfs))
        
        if len(downstream_tfs) != 0:
            
            average_tf_expression = np.zeros((adata.n_obs, ))
            
            for tf in downstream_tfs:
                
                average_tf_expression += adata[:, tf].X.toarray().flatten()
                
            average_tf_expression /= len(downstream_tfs)
            
            inflow_expressions_adjusted[:, i] *= average_tf_expression
            
    inflow_downstream_tfs = []
    for i, inflow_var in enumerate(inflow_vars):
        
        interactions_for_receptor = unique_inflow_vars_and_interactions[inflow_var]
        downstream_tfs = unique_inflow_vars_and_tfs[inflow_var]
        inflow_downstream_tfs.append('_'.join(sorted(downstream_tfs)))
        
    adata_inflow = sc.AnnData(X=inflow_expressions_adjusted)
    adata_inflow.var.index = pd.Index(inflow_vars)
    adata_inflow.var['downstream_tfs'] = inflow_downstream_tfs
    adata_inflow.var['type'] = 'inflow' # Define variable types
    adata_inflow.var['interactions'] = inflow_interactions

    return adata_inflow, inflow_vars

def construct_outflow_signals_cellchat(adata: sc.AnnData,
                                    cellchat_output_key: str, 
                                    ):
    
    cellchat_output_merged = pd.concat([adata.uns[cellchat_output_key][sample] for sample in adata.uns[cellchat_output_key]])
    cellchat_interactions = cellchat_output_merged['interaction_name_2'].unique().tolist()
    outflow_vars = []
    relevant_interactions = {}

    for inter in cellchat_interactions:

        ligand = inter.split(' - ')[0]

        if ligand not in outflow_vars:

            add_ligand = False

            # Sometimes this ligand not is not the actual symbol name (CellChat errors)
            if ligand not in adata.var_names:
                # Get the alternative name for the ligand
                ligand = cellchat_output_merged[cellchat_output_merged['interaction_name_2'] == inter]['ligand'].values[0]

                if (ligand in adata.var_names)&(ligand not in outflow_vars):
                    add_ligand = True
                    

            else:
                if ligand not in outflow_vars:
                    add_ligand = True

            if add_ligand:
                outflow_vars.append(ligand)

                relevant_interactions[ligand] = [inter]

        else:
            interactions_with_ligand = relevant_interactions[ligand]
            interactions_with_ligand.append(inter)
            relevant_interactions[ligand] = interactions_with_ligand

    outflow_expressions = np.zeros((adata.n_obs, len(outflow_vars)))

    for i, signal in enumerate(outflow_vars):
        outflow_expressions[:, i] = adata[:, signal].X.toarray().flatten()

    adata_outflow = sc.AnnData(X=outflow_expressions)
    adata_outflow.var.index = pd.Index(outflow_vars)
    adata_outflow.var['downstream_tfs'] = '' # For housekeeping for later
    adata_outflow.var['type'] = 'outflow' # Define variable types

    relevant_interactions_of_ligands = []
    for ligand in outflow_vars:
        interactions_of_ligand = relevant_interactions[ligand]
        relevant_interactions_of_ligands.append('/'.join(interactions_of_ligand))
        
    adata_outflow.var['interactions'] = relevant_interactions_of_ligands # Define variable types

    return adata_outflow, outflow_vars

def construct_flows_from_cellchat(adata: sc.AnnData,
                                cellchat_output_key: str,
                                gem_expr_key: str = 'X_gem',
                                scale_gem_expr: bool = True,
                                model_organism: str= 'mouse',
                                flowsig_network_key: str = 'flowsig_network',
                                flowsig_expr_key: str = 'X_flow'):


    # Define the expression
    adata_outflow, outflow_vars = construct_outflow_signals_cellchat(adata, cellchat_output_key)

    adata_inflow, inflow_vars = construct_inflow_signals_cellchat(adata, cellchat_output_key, model_organism)

    adata_gem, flow_gem_vars = construct_gem_expressions(adata, gem_expr_key, scale_gem_expr)

    # Determine the flow_variables
    flow_variables = outflow_vars + inflow_vars + flow_gem_vars

    flow_expressions = np.zeros((adata.n_obs, len(flow_variables)))

    for i, outflow_var in enumerate(outflow_vars):
        flow_expressions[:, i] = adata_outflow[:, outflow_var].X.toarray().flatten()

    for i, inflow_var in enumerate(inflow_vars):
        flow_expressions[:, len(outflow_vars) + i] = adata_inflow[:, inflow_var].X.toarray().flatten()

    for i, gem in enumerate(flow_gem_vars):
        flow_expressions[:, len(outflow_vars) + len(inflow_vars) + i] = adata_gem[:, gem].X.flatten()

    flow_variable_types = adata_outflow.var['type'].tolist() \
                            + adata_inflow.var['type'].tolist() \
                            + adata_gem.var['type'].tolist()
    
    flow_downstream_tfs = adata_outflow.var['downstream_tfs'].tolist() \
                            + adata_inflow.var['downstream_tfs'].tolist() \
                            + adata_gem.var['downstream_tfs'].tolist()
    
    flow_interactions = adata_outflow.var['interactions'].tolist() \
                            + adata_inflow.var['interactions'].tolist() \
                            + adata_gem.var['interactions'].tolist()

    # Store the type, relevant downstream_TF, and received interactions for each variable
    # Store all the information on the flow variables
    flow_var_info = pd.DataFrame(index = pd.Index(flow_variables),
                                 data = {'Type': flow_variable_types,
                                      'Downstream_TF': flow_downstream_tfs,
                                      'Interaction': flow_interactions})
    
    adata.obsm[flowsig_expr_key] = flow_expressions
    adata.uns[flowsig_network_key] = {'flow_var_info': flow_var_info}
    
def construct_inflow_signals_cellphonedb(adata: sc.AnnData,
                                    cellphonedb_output_key: str,
                                    cellphonedb_active_tfs_key: str):
    
    ccc_output_merged = pd.concat([adata.uns[cellphonedb_output_key][sample] for sample in adata.uns[cellphonedb_output_key]])
    cpdb_active_tfs_merged = pd.concat([adata.uns[cellphonedb_active_tfs_key][sample] for sample in adata.uns[cellphonedb_active_tfs_key]])
    ccc_interactions = ccc_output_merged['interacting_pair'].unique().tolist()

    unique_inflow_vars_and_interactions = {}

    # The inflow variables are constructed from receptor gene expression
    # that will be weighted by the average expression of their downstream
    # TF targets. These can be multi-units, but because of the weighting,
    # we hestitate to call them receptors in the context of cellular flows.
    for i, interaction in enumerate(ccc_interactions):
        receptors = ccc_output_merged[ccc_output_merged['interacting_pair'] == interaction]['gene_b'].dropna().unique().tolist()
        
        for receptor in receptors:
            if receptor in adata.var_names:
                if receptor not in unique_inflow_vars_and_interactions:
                    unique_inflow_vars_and_interactions[receptor] = [interaction]
                    
                else:
                    interactions_for_receptor = unique_inflow_vars_and_interactions[receptor]
                    interactions_for_receptor.append(interaction)  
                    unique_inflow_vars_and_interactions[receptor] = interactions_for_receptor
                
    inflow_vars = sorted(list(unique_inflow_vars_and_interactions.keys()))
    num_inflow_vars = len(inflow_vars)

    inflow_expressions = np.zeros((adata.n_obs, num_inflow_vars)) 

    for i, receptor in enumerate(inflow_vars):
        
        inflow_expressions[:, i] = adata[:, receptor].X.toarray().flatten()
        
    inflow_expressions_adjusted = inflow_expressions.copy()

    inflow_interactions = []
    unique_inflow_vars_and_tfs = {}

    for i, receptor in enumerate(inflow_vars):
        
        interactions_for_receptor = unique_inflow_vars_and_interactions[receptor]
        inflow_interactions.append('/'.join(sorted(interactions_for_receptor)))
        
        # Go through each category
        possible_downstream_tfs = cpdb_active_tfs_merged[cpdb_active_tfs_merged['gene_b'] == receptor]['active_TF'].dropna().unique().tolist()
                    
        if len(possible_downstream_tfs) != 0:
            downstream_tfs = np.intersect1d(possible_downstream_tfs, adata.var_names)
        
            unique_inflow_vars_and_tfs[receptor] = sorted(list(downstream_tfs))

        else:
            unique_inflow_vars_and_tfs[receptor] = ''
        
        if len(possible_downstream_tfs) != 0:
            
            average_tf_expression = np.zeros((adata.n_obs, ))
            
            for tf in possible_downstream_tfs:
                
                average_tf_expression += adata[:, tf].X.toarray().flatten()
                
            average_tf_expression /= len(possible_downstream_tfs)
            
            inflow_expressions_adjusted[:, i] *= average_tf_expression
            
    inflow_downstream_tfs = []
    for i, inflow_var in enumerate(inflow_vars):
        
        interactions_for_receptor = unique_inflow_vars_and_interactions[inflow_var]
        downstream_tfs = unique_inflow_vars_and_tfs[inflow_var]
        inflow_downstream_tfs.append('_'.join(sorted(downstream_tfs)))
        
    adata_inflow = sc.AnnData(X=inflow_expressions_adjusted)
    adata_inflow.var.index = pd.Index(inflow_vars)
    adata_inflow.var['downstream_tfs'] = inflow_downstream_tfs
    adata_inflow.var['type'] = 'inflow' # Define variable types
    adata_inflow.var['interactions'] = inflow_interactions

    return adata_inflow, inflow_vars

def construct_outflow_signals_cellphonedb(adata: sc.AnnData,
                                    cellphonedb_output_key: str):
    
    cellphonedb_output_merged = pd.concat([adata.uns[cellphonedb_output_key][sample] for sample in adata.uns[cellphonedb_output_key]])
    cellphonedb_interactions = cellphonedb_output_merged['interacting_pair'].unique().tolist()
    outflow_vars = []
    relevant_interactions = {}

    for inter in cellphonedb_interactions:

        ligands = cellphonedb_output_merged[cellphonedb_output_merged['interacting_pair'] == inter]['gene_a'].dropna().unique().tolist()

        for ligand in ligands:
            if (ligand in adata.var_names):


                if (ligand not in outflow_vars):
                    outflow_vars.append(ligand)

                    relevant_interactions[ligand] = [inter]

                else:
                    interactions_with_ligand = relevant_interactions[ligand]
                    interactions_with_ligand.append(inter)
                    relevant_interactions[ligand] = interactions_with_ligand

    outflow_expressions = np.zeros((adata.n_obs, len(outflow_vars)))

    for i, signal in enumerate(outflow_vars):
        outflow_expressions[:, i] = adata[:, signal].X.toarray().flatten()

    adata_outflow = sc.AnnData(X=outflow_expressions)
    adata_outflow.var.index = pd.Index(outflow_vars)
    adata_outflow.var['downstream_tfs'] = '' # For housekeeping for later
    adata_outflow.var['type'] = 'outflow' # Define variable types

    relevant_interactions_of_ligands = []
    for ligand in outflow_vars:
        interactions_of_ligand = relevant_interactions[ligand]
        relevant_interactions_of_ligands.append('/'.join(interactions_of_ligand))
        
    adata_outflow.var['interactions'] = relevant_interactions_of_ligands # Define variable types

    return adata_outflow, outflow_vars

def construct_flows_from_cellphonedb(adata: sc.AnnData,
                                cellphonedb_output_key: str,
                                cellphonedb_tfs_key: str,
                                gem_expr_key: str = 'X_gem',
                                scale_gem_expr: bool = True,
                                model_organism: str = 'human',
                                flowsig_network_key: str = 'flowsig_network',
                                flowsig_expr_key: str = 'X_flow'):

    if model_organism != 'human':
        ValueError("CellPhoneDB only supports human data.")

    # Define the expression
    adata_outflow, outflow_vars = construct_outflow_signals_cellphonedb(adata, cellphonedb_output_key)

    adata_inflow, inflow_vars = construct_inflow_signals_cellphonedb(adata, cellphonedb_output_key, cellphonedb_tfs_key)

    adata_gem, flow_gem_vars = construct_gem_expressions(adata, gem_expr_key, scale_gem_expr)

    # Determine the flow_variables
    flow_variables = outflow_vars + inflow_vars + flow_gem_vars

    flow_expressions = np.zeros((adata.n_obs, len(flow_variables)))

    for i, outflow_var in enumerate(outflow_vars):
        flow_expressions[:, i] = adata_outflow[:, outflow_var].X.toarray().flatten()

    for i, inflow_var in enumerate(inflow_vars):
        flow_expressions[:, len(outflow_vars) + i] = adata_inflow[:, inflow_var].X.toarray().flatten()

    for i, gem in enumerate(flow_gem_vars):
        flow_expressions[:, len(outflow_vars) + len(inflow_vars) + i] = adata_gem[:, gem].X.flatten()

    flow_variable_types = adata_outflow.var['type'].tolist() \
                            + adata_inflow.var['type'].tolist() \
                            + adata_gem.var['type'].tolist()
    
    flow_downstream_tfs = adata_outflow.var['downstream_tfs'].tolist() \
                            + adata_inflow.var['downstream_tfs'].tolist() \
                            + adata_gem.var['downstream_tfs'].tolist()
    
    flow_interactions = adata_outflow.var['interactions'].tolist() \
                            + adata_inflow.var['interactions'].tolist() \
                            + adata_gem.var['interactions'].tolist()

    # Store the type, relevant downstream_TF, and received interactions for each variable
    # Store all the information on the flow variables
    flow_var_info = pd.DataFrame(index = pd.Index(flow_variables),
                                 data = {'Type': flow_variable_types,
                                      'Downstream_TF': flow_downstream_tfs,
                                      'Interaction': flow_interactions})
    
    adata.obsm[flowsig_expr_key] = flow_expressions
    adata.uns[flowsig_network_key] = {'flow_var_info': flow_var_info}
    
def construct_inflow_signals_liana(adata: sc.AnnData,
                                    liana_output_key: str, 
                                    use_tfs: bool = False,
                                    model_organism: str = 'human'):
    
    model_organisms = ['human', 'mouse']

    if model_organism not in model_organisms:
        raise ValueError ("Invalid model organism. Please select one of: %s" % model_organisms)
    
    if use_tfs:
    
        data_dir = os.path.join(os.path.dirname(__file__), '..', 'data')
        data_path = os.path.join(data_dir, 'cellchat_interactions_and_tfs_' + model_organism + '.csv')

        cellchat_interactions_and_tfs = pd.read_csv(data_path, index_col=0)

        ccc_output_merged = pd.concat([adata.uns[liana_output_key][sample] for sample in adata.uns[liana_output_key]])
        inflow_vars = sorted(ccc_output_merged['receptor_complex'].unique().tolist())
        inflow_vars = [var.replace('_', '+') for var in inflow_vars]

        unique_inflow_vars_and_interactions = {}

        # The inflow variables are constructed from receptor gene expression
        # that will be weighted by the average expression of their downstream
        # TF targets. These can be multi-units, but because of the weighting,
        # we hestitate to call them receptors in the context of cellular flows.
        for inflow_var in inflow_vars:
            relevant_rec = inflow_var.replace('+', '_')

            relevant_ligands = sorted(ccc_output_merged[ccc_output_merged['receptor_complex'] == relevant_rec]['ligand_complex'].unique().tolist())
            interactions_for_receptor = []

            for lig in relevant_ligands:
                if '+' in inflow_var:
                    interactions_for_receptor.append(lig + ' - ' + '(' + inflow_var + ')')
                else:
                    interactions_for_receptor.append(lig + ' - ' + inflow_var)

            unique_inflow_vars_and_interactions[inflow_var] = interactions_for_receptor
                
        inflow_vars = sorted(list(unique_inflow_vars_and_interactions.keys()))
        num_inflow_vars = len(inflow_vars)

        inflow_expressions = np.zeros((adata.n_obs, num_inflow_vars)) 

        for i, receptor in enumerate(inflow_vars):
            
            receptor_expression = np.ones((adata.n_obs, ))
            
            split_receptor = receptor.split('+')
            considered_receptors = []
            
            for unit in split_receptor:
                
                if unit not in adata.var_names:
                    print(unit)
                    
                else:
                    
                    unit_expression = adata[:, unit].X.toarray().flatten()

                    receptor_expression *= unit_expression
                    considered_receptors.append(unit)
                    
            if len(considered_receptors) != 0:
                inflow_expressions[:, i] = receptor_expression**(1.0 / len(considered_receptors))
            
        inflow_expressions_adjusted = inflow_expressions.copy()

        inflow_interactions = []
        unique_inflow_vars_and_tfs = {}

        for i, receptor in enumerate(inflow_vars):
            
            interactions_for_receptor = unique_inflow_vars_and_interactions[receptor]
            
            relevant_interactions = cellchat_interactions_and_tfs[cellchat_interactions_and_tfs['interaction_name_2'].isin(interactions_for_receptor)]
            
            joined_interactions = '/'.join(sorted(interactions_for_receptor))
            inflow_interactions.append(joined_interactions)
            
            downstream_tfs = []
            
            # Go through each category
            possible_downstream_tfs = relevant_interactions['Receptor-TF-combined'].dropna().tolist()
            
            for unit in possible_downstream_tfs:
                split_unit = unit.split('_')
                for tf in split_unit:
                    if tf not in downstream_tfs:
                        downstream_tfs.append(tf)
                        
            downstream_tfs = np.intersect1d(downstream_tfs, adata.var_names)
            
            unique_inflow_vars_and_tfs[receptor] = sorted(list(downstream_tfs))
            
            if len(downstream_tfs) != 0:
                
                average_tf_expression = np.zeros((adata.n_obs, ))
                
                for tf in downstream_tfs:
                    
                    average_tf_expression += adata[:, tf].X.toarray().flatten()
                    
                average_tf_expression /= len(downstream_tfs)
                
                inflow_expressions_adjusted[:, i] *= average_tf_expression
                
        inflow_downstream_tfs = []
        for i, inflow_var in enumerate(inflow_vars):
            
            interactions_for_receptor = unique_inflow_vars_and_interactions[inflow_var]
            downstream_tfs = unique_inflow_vars_and_tfs[inflow_var]
            inflow_downstream_tfs.append('_'.join(sorted(downstream_tfs)))
            
        adata_inflow = sc.AnnData(X=inflow_expressions_adjusted)
        adata_inflow.var.index = pd.Index(inflow_vars)
        adata_inflow.var['downstream_tfs'] = inflow_downstream_tfs
        adata_inflow.var['type'] = 'inflow' # Define variable types
        adata_inflow.var['interactions'] = inflow_interactions

        return adata_inflow, inflow_vars
    
def construct_outflow_signals_liana(adata: sc.AnnData,
                                    liana_output_key: str, 
                                    ):
    liana_output_merged = pd.concat([adata.uns[liana_output_key][sample] for sample in adata.uns[liana_output_key]])
   
    outflow_vars = sorted(liana_output_merged['ligand_complex'].unique().tolist())
    outflow_vars = [var.replace('_', '+') for var in outflow_vars]

    relevant_interactions = {}

    for outflow_var in outflow_vars:
            
        relevant_lig = outflow_var.replace('+', '_')

        interactions_with_ligand = []

        relevant_receptors = sorted(liana_output_merged[liana_output_merged['ligand_complex'] == relevant_lig]['receptor_complex'].unique().tolist())

        for rec in relevant_receptors:
            
            inter = outflow_var + ' - ' + rec.replace('_', '+')
            interactions_with_ligand.append(inter)

        relevant_interactions[outflow_var] = interactions_with_ligand

    outflow_expressions = np.zeros((adata.n_obs, len(outflow_vars)))

    for i, ligand in enumerate(outflow_vars):
            
        ligand_expression = np.ones((adata.n_obs, ))

        split_ligand = ligand.split('+')
        considered_ligands = []
        
        for unit in split_ligand:
            
            if unit not in adata.var_names:
                print(unit)
                
            else:
                
                unit_expression = adata[:, unit].X.toarray().flatten()

                ligand_expression *= unit_expression
                considered_ligands.append(unit)
                
        if len(considered_ligands) != 0:
            outflow_expressions[:, i] = ligand_expression**(1.0 / len(considered_ligands))

    adata_outflow = sc.AnnData(X=outflow_expressions)
    adata_outflow.var.index = pd.Index(outflow_vars)
    adata_outflow.var['downstream_tfs'] = '' # For housekeeping for later
    adata_outflow.var['type'] = 'outflow' # Define variable types

    relevant_interactions_of_ligands = []
    for ligand in outflow_vars:
        interactions_of_ligand = relevant_interactions[ligand]
        relevant_interactions_of_ligands.append('/'.join(interactions_of_ligand))
        
    adata_outflow.var['interactions'] = relevant_interactions_of_ligands # Define variable types

    return adata_outflow, outflow_vars
   
def construct_flows_from_liana(adata: sc.AnnData,
                                liana_output_key: str,
                                gem_expr_key: str = 'X_gem',
                                scale_gem_expr: bool = True,
                                use_tfs: bool = False,
                                model_organism: str= 'mouse',
                                flowsig_network_key: str = 'flowsig_network',
                                flowsig_expr_key: str = 'X_flow'):

    # Define the expression
    adata_outflow, outflow_vars = construct_outflow_signals_liana(adata, liana_output_key)

    adata_inflow, inflow_vars = construct_inflow_signals_liana(adata, liana_output_key, use_tfs, model_organism)

    adata_gem, flow_gem_vars = construct_gem_expressions(adata, gem_expr_key, scale_gem_expr)

    # Determine the flow_variables
    flow_variables = outflow_vars + inflow_vars + flow_gem_vars

    flow_expressions = np.zeros((adata.n_obs, len(flow_variables)))

    for i, outflow_var in enumerate(outflow_vars):
        flow_expressions[:, i] = adata_outflow[:, outflow_var].X.toarray().flatten()

    for i, inflow_var in enumerate(inflow_vars):
        flow_expressions[:, len(outflow_vars) + i] = adata_inflow[:, inflow_var].X.toarray().flatten()

    for i, gem in enumerate(flow_gem_vars):
        flow_expressions[:, len(outflow_vars) + len(inflow_vars) + i] = adata_gem[:, gem].X.flatten()

    flow_variable_types = adata_outflow.var['type'].tolist() \
                            + adata_inflow.var['type'].tolist() \
                            + adata_gem.var['type'].tolist()
    
    flow_downstream_tfs = adata_outflow.var['downstream_tfs'].tolist() \
                            + adata_inflow.var['downstream_tfs'].tolist() \
                            + adata_gem.var['downstream_tfs'].tolist()
    
    flow_interactions = adata_outflow.var['interactions'].tolist() \
                            + adata_inflow.var['interactions'].tolist() \
                            + adata_gem.var['interactions'].tolist()

    # Store the type, relevant downstream_TF, and received interactions for each variable
    # Store all the information on the flow variables
    flow_var_info = pd.DataFrame(index = pd.Index(flow_variables),
                                 data = {'Type': flow_variable_types,
                                      'Downstream_TF': flow_downstream_tfs,
                                      'Interaction': flow_interactions})
    
    adata.obsm[flowsig_expr_key] = flow_expressions
    adata.uns[flowsig_network_key] = {'flow_var_info': flow_var_info}

def construct_inflow_signals_commot(adata: sc.AnnData,
                                    commot_output_key: str):

    # Inflow variables are inferred from outflow variables
    outflow_vars = sorted(adata.uns[commot_output_key + '-info']['df_ligrec']['ligand'].unique().tolist())
    inflow_vars = ['inflow-' + outflow_var for outflow_var in outflow_vars]


    inflow_interactions = []
    inflow_expressions = np.zeros((adata.n_obs, len(inflow_vars)))
    for i, inflow_var in enumerate(inflow_vars):
        lig = inflow_var.strip('inflow-')
        inferred_interactions = [pair.replace('r-', '') for pair in adata.obsm[commot_output_key + '-sum-receiver'].columns if pair.startswith('r-' + lig)]
        inflow_interactions.append('/'.join(sorted(inferred_interactions)))

        # We sum the total received signal across each interaction at each spot
        for inter in inferred_interactions:
            inflow_expressions[:, i] += adata.obsm[commot_output_key + '-sum-receiver']['r-' + inter]

    adata_inflow = sc.AnnData(X=inflow_expressions)
    adata_inflow.var.index = pd.Index(inflow_vars)
    adata_inflow.var['downstream_tfs'] = ''
    adata_inflow.var['type'] = 'inflow' 
    adata_inflow.var['interactions'] = inflow_interactions

    return adata_inflow, inflow_vars

def construct_outflow_signals_commot(adata: sc.AnnData,
                                    commot_output_key: str, 
                                    ):
    
    # Inflow variables are inferred from outflow variables
    outflow_vars = sorted(adata.uns[commot_output_key + '-info']['df_ligrec']['ligand'].unique().tolist())
    outflow_vars = [var for var in outflow_vars if var in adata.var_names]

    outflow_interactions = []
    outflow_expressions = np.zeros((adata.n_obs, len(outflow_vars)))

    for i, outflow_var in enumerate(outflow_vars):

        inferred_interactions = [pair[2:] for pair in adata.obsm[commot_output_key + '-sum-sender'].columns if pair.startswith('s-' + outflow_var)]
        outflow_interactions.append('/'.join(sorted(inferred_interactions)))

        # Outflow signal expression is simply ligand gene expression
        outflow_expressions[:, i] += adata[:, outflow_var].X.toarray().flatten()

    adata_outflow = sc.AnnData(X=outflow_expressions)
    adata_outflow.var.index = pd.Index(outflow_vars)
    adata_outflow.var['downstream_tfs'] = ''
    adata_outflow.var['type'] = 'outflow' 
    adata_outflow.var['interactions'] = outflow_interactions

    return adata_outflow, outflow_vars

def construct_flows_from_commot(adata: sc.AnnData,
                                commot_output_key: str,
                                gem_expr_key: str = 'X_gem',
                                scale_gem_expr: bool = True,
                                flowsig_network_key: str = 'flowsig_network',
                                flowsig_expr_key: str = 'X_flow'):
    
    # Define the expression
    adata_outflow, outflow_vars = construct_outflow_signals_commot(adata, commot_output_key)

    adata_inflow, inflow_vars = construct_inflow_signals_commot(adata, commot_output_key)

    adata_gem, flow_gem_vars = construct_gem_expressions(adata, gem_expr_key, scale_gem_expr)

    # Determine the flow_variables
    flow_variables = outflow_vars + inflow_vars + flow_gem_vars

    flow_expressions = np.zeros((adata.n_obs, len(flow_variables)))

    for i, outflow_var in enumerate(outflow_vars):
        flow_expressions[:, i] = adata_outflow[:, outflow_var].X.toarray().flatten()

    for i, inflow_var in enumerate(inflow_vars):
        flow_expressions[:, len(outflow_vars) + i] = adata_inflow[:, inflow_var].X.toarray().flatten()

    for i, gem in enumerate(flow_gem_vars):
        flow_expressions[:, len(outflow_vars) + len(inflow_vars) + i] = adata_gem[:, gem].X.flatten()

    flow_variable_types = adata_outflow.var['type'].tolist() \
                            + adata_inflow.var['type'].tolist() \
                            + adata_gem.var['type'].tolist()
    
    flow_downstream_tfs = adata_outflow.var['downstream_tfs'].tolist() \
                            + adata_inflow.var['downstream_tfs'].tolist() \
                            + adata_gem.var['downstream_tfs'].tolist()
    
    flow_interactions = adata_outflow.var['interactions'].tolist() \
                            + adata_inflow.var['interactions'].tolist() \
                            + adata_gem.var['interactions'].tolist()

    # Store the type, relevant downstream_TF, and received interactions for each variable
    # Store all the information on the flow variables
    flow_var_info = pd.DataFrame(index = pd.Index(flow_variables),
                                 data = {'Type': flow_variable_types,
                                      'Downstream_TF': flow_downstream_tfs,
                                      'Interaction': flow_interactions})
    
    adata.obsm[flowsig_expr_key] = flow_expressions
    adata.uns[flowsig_network_key] = {'flow_var_info': flow_var_info}




# __init__ file
import warnings
warnings.filterwarnings('ignore')

from . import preprocessing as pp
from . import tools as tl
from . import plotting as pl
from . import utilities as ul

#import sys
#sys.modules.update({f'{__name__}.{m}': globals()[m] for m in ['pp','tl','pl', 'ul']})


import numpy as np
from kneed import DataGenerator, KneeLocator
from .dp_related import dp_bucketized
import torch
import matplotlib.pyplot as plt

def plot_ll_curve(gaston_model, A, S, max_domain_num=8, start_from=2):
    ll_list=get_ll_list(gaston_model, A, S, num_buckets=150, kmax=max_domain_num)
    
    kneedle=KneeLocator(np.arange(start_from,len(ll_list)+1), ll_list[start_from-1:], curve="convex", direction="decreasing")
    kneedle_opt=kneedle.knee
    print(f'Kneedle number of domains: {kneedle_opt}')

    fig,ax=plt.subplots(figsize=(4,3))
    
    plt.plot(np.arange(start_from,len(ll_list)+1), ll_list[start_from-1:])
    plt.scatter(np.arange(start_from,len(ll_list)+1), ll_list[start_from-1:])
    plt.axvline(kneedle_opt,ls='--',color='grey')
    
    plt.xlabel('Number of domains',fontsize=17)
    plt.ylabel('Negative \nlog-likelihood',fontsize=17)
    ax.tick_params(axis='both', which='major', labelsize=15)
    plt.xticks(np.arange(start_from,len(ll_list)+1),fontsize=15)
    
    plt.tight_layout()

def get_ll_list(model, A, S, num_buckets=150, kmax=10):
    N=A.shape[0]
    S_torch=torch.Tensor(S)
    belayer_depth=model.spatial_embedding(torch.Tensor(S)).detach().numpy().flatten()
    
    bin_endpoints=np.linspace(np.min(belayer_depth),np.max(belayer_depth)+0.01,num_buckets+1)
    error_mat,seg_map=dp_bucketized(A.T, bin_endpoints, kmax, xcoords=belayer_depth)
    return error_mat[-1,:]


from .model.model import MulanConfig, scMulanModel
from .model.model_kvcache import scMulanModel_kv
from .utils.hf_tokenizer import scMulanTokenizer
from .scMulan import model_inference
from .scMulan_npu import model_inference_npu
from .reference.GeneSymbolUniform.pyGSUni import  GeneSymbolUniform
from .utils.utils import cell_type_smoothing, visualize_selected_cell_types


import pandas as pd
import scanpy as sc
from sklearn.cluster import KMeans

def construct_spatial_blocks(adata: sc.AnnData,
                             n_blocks: int,
                             use_graph: bool = False,
                             graph_adjacency: str = 'spatial_connectivities',
                             resolution: float = None,
                             spatial_block_key: str = "spatial_block",
                             spatial_key: str = "spatial"):
    
    # If we want to construct spatial clusters from the graph directly, we use leiden clustering
    if use_graph: 

        if resolution is None:
            ValueError('Need to specify a clustering resolution to construct blocks from spatial graph.')

        else:
            sc.tl.leiden(adata, resolution=resolution, key_added=spatial_block_key, adjacency=graph_adjacency)

    else:         # Run k-means clustering on the spatial coordinates (produces more "even" blocks)

        kmeans = KMeans(n_clusters=n_blocks).fit(adata.obsm[spatial_key])
        adata.obs[spatial_block_key] = pd.Series(kmeans.labels_, dtype='category').values


import numpy as np
import scipy.sparse as sp
import torch
import torch.utils.data as data_utils


class EdgeSampler(data_utils.Dataset):
    """Sample edges and non-edges uniformly from a graph.

    Args:
        A: adjacency matrix.
        num_pos: number of edges per batch.
        num_neg: number of non-edges per batch.
    """
    def __init__(self, A, num_pos=1000, num_neg=1000):
        self.num_pos = num_pos
        self.num_neg = num_neg
        self.A = A
        self.edges = np.transpose(A.nonzero())
        self.num_nodes = A.shape[0]
        self.num_edges = self.edges.shape[0]

    def __getitem__(self, key):
        np.random.seed(key)
        edges_idx = np.random.randint(0, self.num_edges, size=self.num_pos, dtype=np.int64)
        next_edges = self.edges[edges_idx, :]

        # Select num_neg non-edges
        generated = False
        while not generated:
            candidate_ne = np.random.randint(0, self.num_nodes, size=(2*self.num_neg, 2), dtype=np.int64)
            cne1, cne2 = candidate_ne[:, 0], candidate_ne[:, 1]
            to_keep = (1 - self.A[cne1, cne2]).astype(np.bool).A1 * (cne1 != cne2)
            next_nonedges = candidate_ne[to_keep][:self.num_neg]
            generated = to_keep.sum() >= self.num_neg
        return torch.LongTensor(next_edges), torch.LongTensor(next_nonedges)

    def __len__(self):
        return 2**32

def collate_fn(batch):
    edges, nonedges = batch[0]
    return (edges, nonedges)

def get_edge_sampler(A, num_pos=1000, num_neg=1000, num_workers=2):
    data_source = EdgeSampler(A, num_pos, num_neg)
    return data_utils.DataLoader(data_source, num_workers=num_workers, collate_fn=collate_fn)



from ._plotting import plot_cell_communication
from ._plotting import plot_cluster_communication_network
from ._plotting import plot_cluster_communication_dotplot
from ._plotting import plot_communication_dependent_genes
from ._plotting import plot_communication_impact


from .basic_distributions import Distribution
from .bernoulli import Bernoulli
from .univariate_gaussian import UnivariateGaussian
from .bernoulli_gaussian import BernoulliGaussian
from .beta import Beta
from .binomial import Binomial
from .gamma import Gamma
from .multivariate_gaussian import MultivariateGaussian
from .poisson import Poisson
# from .multi_task_GP import MyMultitaskGPModel, MultitaskGPModel, _myGaussianLikelihoodBase, myExactGP


import numpy as np
import matplotlib.pyplot as plt
import pandas as pd

from sklearn.preprocessing import normalize

from mpl_toolkits.axes_grid1.inset_locator import mark_inset, inset_axes

import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.utils
import torch.distributions
from .dp_related import rotate_by_theta

def plot_clusters(gaston_labels, S, fig=None, ax=None, figsize=(5,8), colors=None, color_palette=plt.cm.Dark2, s=20,labels=None,
                 lgd=False, rotate=None, show_boundary=False, gaston_isodepth=None, boundary_lw=10,bbox_to_anchor=None,
                 linear_transform=None):

    if rotate is not None:
        S=rotate_by_theta(S,rotate)
    elif linear_transform is not None:
        S=(linear_transform @ S.T).T
    
    
    if fig is None or ax is None:
        fig,ax=plt.subplots(figsize=figsize)

    if colors is None:
        colors=np.array([color_palette(i) for i in range(len(np.unique(gaston_labels)))])

    for i,t in enumerate(np.unique(gaston_labels)):
        pts_t=np.where(gaston_labels==t)[0]
        if labels is not None:
            l=labels[i]
        else:
            l=i
        ax.scatter(S[pts_t,0], S[pts_t,1],s=s, color=colors[int(t)], label=l)

    if show_boundary:
        plt.tricontour(S[:,0], S[:,1], gaston_isodepth, 
               levels=[np.min(gaston_isodepth[gaston_labels==i]) for i in range(len(np.unique(gaston_labels)))[1:]], 
               linewidths=boundary_lw, colors='k', linestyles='-')
    
    plt.axis('off')
    if lgd:
        plt.legend(bbox_to_anchor=bbox_to_anchor)
    
#######

def plot_clusters_restrict(gaston_labels, S, gaston_isodepth, isodepth_min=0, isodepth_max=1, restrict_color='lime', fig=None, ax=None, figsize=(5,8), colors=None, color_palette=plt.cm.Dark2, s=20,labels=None,lgd=False, rotate=None):

    if rotate is not None:
        S=rotate_by_theta(S,rotate)
    
    if fig is None or ax is None:
        fig,ax=plt.subplots(figsize=figsize)

    if colors is None:
        colors=np.array([color_palette(i) for i in range(len(np.unique(gaston_labels)))])

    for i,t in enumerate(np.unique(gaston_labels)):
        pts_t=np.where(gaston_labels==t)[0]
        if labels is not None:
            l=labels[i]
        else:
            l=i
        ax.scatter(S[pts_t,0], S[pts_t,1],s=s, color=colors[int(t)], label=l)
    plt.axis('off')
    if lgd:
        plt.legend()

    locs=np.array( [i for i in range(len(gaston_isodepth)) if isodepth_min < gaston_isodepth[i] < isodepth_max] )
    plt.scatter(S[locs,0],S[locs,1],c=restrict_color,alpha=0.5)

#######

    
def plot_isodepth(gaston_isodepth, S, mod, figsize=(5,8), contours=True, 
                  contour_levels=4, contour_lw=1, contour_fs=10, colorbar=True,
                  s=20,cbar_fs=10, axis_off=True, streamlines=False, streamlines_lw=1.5, 
                  rotate=None, cmap='coolwarm', norm=None,
                 arrowsize=2, neg_gradient=False, scaling_factors=None, 
                 gaston_labels_for_scaling=None, linear_transform=None,
                 n_neighbors=1000):
    
    if rotate is not None:
        S_rotated=rotate_by_theta(S,rotate)
    elif linear_transform is not None:
        S_rotated=(linear_transform @ S.T).T
    else:
        S_rotated=S
    
    fig,ax=plt.subplots(figsize=figsize)

    im1=ax.scatter(S_rotated[:,0], S_rotated[:,1], c=gaston_isodepth, cmap=cmap, s=s, norm=norm)
    if axis_off:
        plt.axis('off')

    if contours:
        CS=ax.tricontour(S_rotated[:,0], S_rotated[:,1], gaston_isodepth, levels=contour_levels, linewidths=contour_lw, colors='k', linestyles='solid')
        ax.clabel(CS, CS.levels, inline=True, fontsize=contour_fs)
    if colorbar:
        cbar=plt.colorbar(im1)
        cbar.ax.tick_params(labelsize=cbar_fs)
    if streamlines:
        x=torch.tensor(S,requires_grad=True).float()
        G=torch.autograd.grad(outputs=mod.spatial_embedding(x).flatten(),inputs=x, grad_outputs=torch.ones_like(x[:,0]))[0]
        G=G.detach().numpy()
        if neg_gradient:
            G=-1*G

        if scaling_factors is not None:
            L=len(np.unique(gaston_labels_for_scaling))
            for l in range(L):
                G[gaston_labels_for_scaling==l,:] = scaling_factors[l] * G[gaston_labels_for_scaling==l,:]
        
        if rotate is not None:
            G_rotated=rotate_by_theta(G,rotate)
        elif linear_transform is not None:
            G_rotated=(linear_transform @ G.T).T
        else:
            G_rotated=G
        
        # CODE FROM scVelo
        smooth=None
        min_mass=None
        
        cutoff_perc=0
        
        X_grid, V_grid = compute_velocity_on_grid(
                    X_emb=S_rotated,
                    V_emb=G_rotated,
                    density=1,
                    smooth=smooth,
                    min_mass=min_mass,
                    n_neighbors=n_neighbors,
                    adjust_for_stream=True,
                    cutoff_perc=cutoff_perc,
                )
        lengths = np.sqrt((V_grid**2).sum(0))
        linewidth=streamlines_lw
        linewidth *= 2 * lengths / lengths[~np.isnan(lengths)].max()
        
        
        density=1
        stream_kwargs = {
                "linewidth": linewidth,
                "density": density,
                "zorder": 3,
                "color": "k",
                "arrowsize": arrowsize,
                "arrowstyle": "-|>",
                "maxlength": 1000,
                "integration_direction": "both",
            }
        
        ax.streamplot(X_grid[0], X_grid[1], V_grid[0], V_grid[1], **stream_kwargs)
        plt.axis('off')

#######################################################

# streamlines code from

from sklearn.neighbors import NearestNeighbors
from scipy.stats import norm as normal

def compute_velocity_on_grid(
    X_emb,
    V_emb,
    density=None,
    smooth=None,
    n_neighbors=None,
    min_mass=None,
    adjust_for_stream=False,
    cutoff_perc=None,
):
    """TODO."""
    # remove invalid cells
    idx_valid = np.isfinite(X_emb.sum(1) + V_emb.sum(1))
    X_emb = X_emb[idx_valid]
    V_emb = V_emb[idx_valid]

    # prepare grid
    n_obs, n_dim = X_emb.shape
    density = 1 if density is None else density
    smooth = 0.5 if smooth is None else smooth

    grs = []
    for dim_i in range(n_dim):
        m, M = np.min(X_emb[:, dim_i]), np.max(X_emb[:, dim_i])
        m = m - 0.01 * np.abs(M - m)
        M = M + 0.01 * np.abs(M - m)
        gr = np.linspace(m, M, int(50 * density))
        grs.append(gr)

    meshes_tuple = np.meshgrid(*grs)
    X_grid = np.vstack([i.flat for i in meshes_tuple]).T

    # estimate grid velocities
    if n_neighbors is None:
        n_neighbors = int(n_obs / 50)
    nn = NearestNeighbors(n_neighbors=n_neighbors, n_jobs=-1)
    nn.fit(X_emb)
    dists, neighs = nn.kneighbors(X_grid)

    scale = np.mean([(g[1] - g[0]) for g in grs]) * smooth
    weight = normal.pdf(x=dists, scale=scale)
    p_mass = weight.sum(1)

    V_grid = (V_emb[neighs] * weight[:, :, None]).sum(1)
    V_grid /= np.maximum(1, p_mass)[:, None]
    if min_mass is None:
        min_mass = 1

    if adjust_for_stream:
        X_grid = np.stack([np.unique(X_grid[:, 0]), np.unique(X_grid[:, 1])])
        ns = int(np.sqrt(len(V_grid[:, 0])))
        V_grid = V_grid.T.reshape(2, ns, ns)

        mass = np.sqrt((V_grid**2).sum(0))
        min_mass = 10 ** (min_mass - 6)  # default min_mass = 1e-5
        min_mass = np.clip(min_mass, None, np.max(mass) * 0.9)
        cutoff = mass.reshape(V_grid[0].shape) < min_mass

        if cutoff_perc is None:
            cutoff_perc = 5
        length = np.sum(np.mean(np.abs(V_emb[neighs]), axis=1), axis=1).T
        length = length.reshape(ns, ns)
        cutoff |= length < np.percentile(length, cutoff_perc)

        V_grid[0][cutoff] = np.nan
    else:
        min_mass *= np.percentile(p_mass, 99) / 100
        X_grid, V_grid = X_grid[p_mass > min_mass], V_grid[p_mass > min_mass]

    return X_grid, V_grid


from __future__ import annotations

import warnings
from typing import TYPE_CHECKING

import numpy as np
from sklearn.utils import check_array, check_random_state

from scanpy import logging as logg
from scanpy._compat import old_positionals
from scanpy._settings import settings
from scanpy._utils import NeighborsView
from scanpy.tools._utils import _choose_representation, get_init_pos_from_paga
#from .._settings import EMOJI

EMOJI = {
    "start": "🚀",
    "done": "✅",
    "error": "❌",
    "warning": "⚠️",
    "info": "ℹ️",
}

if TYPE_CHECKING:
    from typing import Literal

    from anndata import AnnData

    from scanpy._utils.random import _LegacyRandom

    _InitPos = Literal["paga", "spectral", "random"]


@old_positionals(
    "min_dist",
    "spread",
    "n_components",
    "maxiter",
    "alpha",
    "gamma",
    "negative_sample_rate",
    "init_pos",
    "random_state",
    "a",
    "b",
    "copy",
    "method",
    "neighbors_key",
)
def umap(  # noqa: PLR0913, PLR0915
    adata: AnnData,
    *,
    min_dist: float = 0.5,
    spread: float = 1.0,
    n_components: int = 2,
    maxiter: int | None = None,
    alpha: float = 1.0,
    gamma: float = 1.0,
    negative_sample_rate: int = 5,
    init_pos: _InitPos | np.ndarray | None = "spectral",
    random_state: _LegacyRandom = 0,
    a: float | None = None,
    b: float | None = None,
    method: Literal["umap", "rapids","torchdr","mde"] = "umap",
    key_added: str | None = None,
    neighbors_key: str = "neighbors",
    copy: bool = False,
) -> AnnData | None:
    r"""Embed the neighborhood graph using UMAP with multiple backend options.

    UMAP (Uniform Manifold Approximation and Projection) implementation with support
    for multiple computational backends including GPU-accelerated variants using
    TorchDR and MDE for improved performance on large datasets.

    Arguments:
        adata: Annotated data matrix of shape n_obs × n_vars
        min_dist (float): Minimum distance between embedded points (default: 0.5)
        spread (float): Scale at which embedded points are spread out (default: 1.0)
        n_components (int): Number of dimensions for the embedding (default: 2)
        maxiter (int): Number of optimization iterations (default: None)
        alpha (float): Initial learning rate for optimization (default: 1.0)
        gamma (float): Weighting for negative samples (default: 1.0)
        negative_sample_rate (int): Number of negative samples per positive sample (default: 5)
        init_pos: Initialization method for embedding positions (default: 'spectral')
        random_state (int): Random seed for reproducible results (default: 0)
        a (float): UMAP curve parameter (default: None)
        b (float): UMAP curve parameter (default: None)
        method (str): Implementation method: 'umap', 'torchdr', 'mde', 'rapids' (default: 'umap')
        key_added (str): Key for storing results in AnnData (default: None)
        neighbors_key (str): Key for accessing neighbor information (default: 'neighbors')
        copy (bool): Return copy instead of modifying in place (default: False)

    Returns:
        adata with UMAP embedding if copy=False, otherwise returns modified copy

    """
    adata = adata.copy() if copy else adata

    key_obsm, key_uns = ("X_umap", "umap") if key_added is None else [key_added] * 2

    if neighbors_key is None:  # backwards compat
        neighbors_key = "neighbors"
    if neighbors_key not in adata.uns:
        msg = f"Did not find .uns[{neighbors_key!r}]. Run `sc.pp.neighbors` first."
        raise ValueError(msg)

    start = logg.info(f"computing UMAP{EMOJI['start']}")

    neighbors = NeighborsView(adata, neighbors_key)

    if "params" not in neighbors or neighbors["params"]["method"] != "umap":
        logg.warning(
            f'.obsp["{neighbors["connectivities_key"]}"] have not been computed using umap'
        )

    with warnings.catch_warnings():
        # umap 0.5.0
        warnings.filterwarnings("ignore", message=r"Tensorflow not installed")
        import umap

    from umap.umap_ import find_ab_params, simplicial_set_embedding

    if a is None or b is None:
        a, b = find_ab_params(spread, min_dist)
    adata.uns[key_uns] = dict(params=dict(a=a, b=b))
    if isinstance(init_pos, str) and init_pos in adata.obsm:
        init_coords = adata.obsm[init_pos]
    elif isinstance(init_pos, str) and init_pos == "paga":
        init_coords = get_init_pos_from_paga(
            adata, random_state=random_state, neighbors_key=neighbors_key
        )
    else:
        init_coords = init_pos  # Let umap handle it
    if hasattr(init_coords, "dtype"):
        init_coords = check_array(init_coords, dtype=np.float32, accept_sparse=False)

    if random_state != 0:
        adata.uns[key_uns]["params"]["random_state"] = random_state
    #random_state = check_random_state(random_state)

    neigh_params = neighbors["params"]
    X = _choose_representation(
        adata,
        use_rep=neigh_params.get("use_rep", None),
        n_pcs=neigh_params.get("n_pcs", None),
        silent=True,
    )
    if method == "umap":
        # the data matrix X is really only used for determining the number of connected components
        # for the init condition in the UMAP embedding
        default_epochs = 500 if neighbors["connectivities"].shape[0] <= 10000 else 200
        n_epochs = default_epochs if maxiter is None else maxiter
        X_umap, _ = simplicial_set_embedding(
            data=X,
            graph=neighbors["connectivities"].tocoo(),
            n_components=n_components,
            initial_alpha=alpha,
            a=a,
            b=b,
            gamma=gamma,
            negative_sample_rate=negative_sample_rate,
            n_epochs=n_epochs,
            init=init_coords,
            random_state=random_state,
            metric=neigh_params.get("metric", "euclidean"),
            metric_kwds=neigh_params.get("metric_kwds", {}),
            densmap=False,
            densmap_kwds={},
            output_dens=False,
            verbose=settings.verbosity > 3,
        )
    elif method == "torchdr":
        from torchdr import UMAP
        import torch
        device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
        n_neighbors = neighbors["params"]["n_neighbors"]
        n_epochs = (
            500 if maxiter is None else maxiter
        )  # 0 is not a valid value for rapids, unlike original umap
        X_contiguous = np.ascontiguousarray(X, dtype=np.float32)
        
        umap = UMAP(
            n_neighbors=n_neighbors,
            n_components=n_components,
            max_iter=n_epochs,
            min_dist=min_dist,
            spread=spread,
            a=a,
            b=b,
            lr = alpha,
            device=device,
            random_state=random_state,
        )
        X_umap = umap.fit(X_contiguous).embedding_.detach().cpu().numpy()

        
        del umap
        import gc
        torch.cuda.empty_cache()
        gc.collect()

    elif method == "mde":
        try:
            from pymde import MDE, constraints, penalties, preprocess
            import torch
        except ImportError as err:
            raise ImportError("Please install pymde package via `pip install pymde`") from err
            
        device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
        
        # Convert random_state to torch seed
        import pymde
        pymde.seed(random_state)
        
        # Get neighbor graph from connectivities
        connectivities = neighbors["connectivities"]
        
        # Convert to PyTorch sparse tensor
        rows, cols = connectivities.nonzero()
        edges = torch.tensor(np.vstack([rows, cols]).T, dtype=torch.int64)
        
        # Filter out self-loops and ensure each edge appears only once
        mask = edges[:, 0] != edges[:, 1]
        edges = edges[mask]
        
        # Sort edges to ensure i < j
        sorted_edges = torch.zeros_like(edges)
        sorted_edges[:, 0] = torch.min(edges[:, 0], edges[:, 1])
        sorted_edges[:, 1] = torch.max(edges[:, 0], edges[:, 1])
        edges = torch.unique(sorted_edges, dim=0)
        
        # Create weights for edges based on connectivities
        weights = torch.ones(edges.shape[0], device=device)
        
        # Add dissimilar edges for better embedding
        n_similar = edges.shape[0]
        n_dissimilar = int(n_similar * min_dist)  # Use min_dist to scale number of dissimilar edges
        
        try:
            # Try to generate dissimilar edges
            dissimilar_edges = preprocess.dissimilar_edges(
                X.shape[0], 
                num_edges=int(n_dissimilar), 
                similar_edges=edges,
            )
            
            # Combine similar and dissimilar edges
            all_edges = torch.cat([edges, dissimilar_edges.to(edges.device)])
            
            # Create weights: positive for similar edges, negative for dissimilar
            all_weights = torch.cat([
                weights,  # Positive weights for similar edges
                -0.5 * torch.ones(dissimilar_edges.shape[0], device=device)  # Negative weights for dissimilar
            ])
            
            # Create PushAndPull distortion function
            distortion_function = penalties.PushAndPull(
                weights=all_weights,
                attractive_penalty=penalties.Log1p,
                repulsive_penalty=penalties.Log,
            )
            
            # Use the combined edges
            edges_to_use = all_edges
            
        except Exception as e:
            # Fallback to simple quadratic penalty if dissimilar edges fail
            logg.warning(f"Failed to generate dissimilar edges: {str(e)}. Using quadratic penalty instead.")
            distortion_function = penalties.Quadratic(weights)
            edges_to_use = edges
        
        n_epochs = 500 if maxiter is None else maxiter
        
        # Create MDE problem with appropriate penalty
        mde = MDE(
            n_items=X.shape[0],
            embedding_dim=n_components,
            edges=edges_to_use.to(device),
            distortion_function=distortion_function,
            constraint=constraints.Standardized(),
            device=device
        )
        
        # Compute embedding with initial random state
        if isinstance(init_coords, np.ndarray):
            embedding = mde.embed(
                verbose=settings.verbosity > 3,
                max_iter=n_epochs,
                # Set random initial state if not using existing embedding
                X=None if init_coords is None else init_coords
            )
        else:
            embedding = mde.embed(
                verbose=settings.verbosity > 3,
                max_iter=n_epochs,
            )
        
        X_umap = embedding.cpu().numpy()
        
        # Clean up
        del mde, embedding, edges, weights
        if 'all_edges' in locals():
            del all_edges
        if 'all_weights' in locals():
            del all_weights
        if 'distortion_function' in locals():
            del distortion_function
    
        
        torch.cuda.empty_cache()
        import gc
        gc.collect()
        
    elif method == "rapids":
        msg = (
            "`method='rapids'` is deprecated. "
            "Use `rapids_singlecell.tl.louvain` instead."
        )
        warnings.warn(msg, FutureWarning, stacklevel=2)
        metric = neigh_params.get("metric", "euclidean")
        if metric != "euclidean":
            msg = (
                f"`sc.pp.neighbors` was called with `metric` {metric!r}, "
                "but umap `method` 'rapids' only supports the 'euclidean' metric."
            )
            raise ValueError(msg)
        from cuml import UMAP

        n_neighbors = neighbors["params"]["n_neighbors"]
        n_epochs = (
            500 if maxiter is None else maxiter
        )  # 0 is not a valid value for rapids, unlike original umap
        X_contiguous = np.ascontiguousarray(X, dtype=np.float32)
        umap = UMAP(
            n_neighbors=n_neighbors,
            n_components=n_components,
            n_epochs=n_epochs,
            learning_rate=alpha,
            init=init_pos,
            min_dist=min_dist,
            spread=spread,
            negative_sample_rate=negative_sample_rate,
            a=a,
            b=b,
            verbose=settings.verbosity > 3,
            random_state=random_state,
        )
        X_umap = umap.fit_transform(X_contiguous)
    adata.obsm[key_obsm] = X_umap  # annotate samples with UMAP coordinates
    logg.info(
        f"    finished {EMOJI['done']}",
        time=start,
        deep=(
            "added\n"
            f"    {key_obsm!r}, UMAP coordinates (adata.obsm)\n"
            f"    {key_uns!r}, UMAP parameters (adata.uns)"
        ),
    )
    return adata if copy else None






from typing import Optional, Union
import sys
import anndata
import numpy as np
import pandas as pd
import scanpy as sc
import networkx as nx
from matplotlib import cm
import matplotlib.pyplot as plt
import plotly
from scipy import sparse
from scipy.spatial import distance_matrix
from scipy.stats import spearmanr, pearsonr
from sklearn.preprocessing import normalize
from sklearn.neighbors import kneighbors_graph
from sklearn.decomposition import PCA


from .._utils import partial_corr
from .._utils import semipartial_corr
from .._utils import treebased_score
from .._utils import treebased_score_multifeature
from .._utils import d_graph_local_jaccard
from .._utils import d_graph_local_jaccard_weighted
from .._utils import d_graph_global_structure
from .._utils import leiden_clustering
from .._utils import moranI_vector_global
from .._utils import preprocess_vector_field
from .._utils import binarize_sparse_matrix

def communication_deg_detection(
    adata: anndata.AnnData,
    n_var_genes: int = None,
    var_genes = None,
    database_name: str = None,
    pathway_name: str = None,
    summary: str = 'receiver',
    lr_pair: tuple = ('total','total'),
    nknots: int = 6,
    n_deg_genes: int = None,
    n_points: int = 50,
    deg_pvalue_cutoff: float = 0.05,
):
    """
    Identify signaling dependent genes

    This function depends on tradeSeq [Van_den_Berge2020]_. Currently, tradeSeq version 1.0.1 with R version 3.6.3 has been tested to work.
    For the R-python interface, rpy2==3.4.2 and anndata2ri==1.0.6 have been tested to work.

    Here, the total received or sent signal for the spots are considered as a "gene expression" where tradeSeq is used to find the correlated genes.

    Parameters
    ----------
    adata
        The data matrix of shape ``n_obs`` × ``n_var``.
        Rows correspond to cells or positions and columns to genes.
        The count data should be available through adata.layers['count'].
        For example, when examining the received signal through the ligand-receptor pair "ligA" and "RecA" infered with the LR database "databaseX", 
        the signaling inference result should be available in 
        ``adata.obsm['commot-databaseX-sum-receiver']['r-ligA-recA']``
    n_var_genes
        The number of most variable genes to test.
    var_genes
        The genes to test. n_var_genes will be ignored if given.
    n_deg_genes
        The number of top deg genes to evaluate yhat.
    pathway_name
        Name of the signaling pathway (choose from the third column of ``.uns['commot-databaseX-info']['df_ligrec']``).
        If ``pathway_name`` is specified, ``lr_pair`` will be ignored.
    summary
        'sender' or 'receiver'
    lr_pair
        A tuple of the ligand-receptor pair.
        If ``pathway_name`` is specified, ``lr_pair`` will be ignored.
    nknots
        Number of knots in spline when constructing GAM.
    n_points
        Number of points on which to evaluate the fitted GAM 
        for downstream clustering and visualization.
    deg_pvalue_cutoff
        The p-value cutoff of genes for obtaining the fitted gene expression patterns.

    Returns
    -------
    df_deg: pd.DataFrame
        A data frame of deg analysis results, including Wald statistics, degree of freedom, and p-value.
    df_yhat: pd.DataFrame
        A data frame of smoothed gene expression values.
    
    References
    ----------

    .. [Van_den_Berge2020] Van den Berge, K., Roux de Bézieux, H., Street, K., Saelens, W., Cannoodt, R., Saeys, Y., ... & Clement, L. (2020). 
        Trajectory-based differential expression analysis for single-cell sequencing data. Nature communications, 11(1), 1-13.

    """
    # setup R environment
    # !!! anndata2ri works only with 3.6.3 on the tested machine
    import rpy2
    import anndata2ri
    import rpy2.robjects as ro
    from rpy2.robjects.conversion import localconverter
    import rpy2.rinterface_lib.callbacks
    import logging
    rpy2.rinterface_lib.callbacks.logger.setLevel(logging.ERROR)

    ro.r('library(tradeSeq)')
    ro.r('library(clusterExperiment)')
    anndata2ri.activate()
    ro.numpy2ri.activate()
    ro.pandas2ri.activate()

    # prepare input adata for R
    adata_deg = anndata.AnnData(
        X = adata.layers['counts'],
        var = pd.DataFrame(index=list(adata.var_names)),
        obs = pd.DataFrame(index=list(adata.obs_names)))
    adata_deg_var = adata_deg.copy()
    sc.pp.filter_genes(adata_deg_var, min_cells=3)
    sc.pp.filter_genes(adata_deg, min_cells=3)
    sc.pp.normalize_total(adata_deg_var, target_sum=1e4)
    sc.pp.log1p(adata_deg_var)
    if n_var_genes is None:
        sc.pp.highly_variable_genes(adata_deg_var, min_mean=0.0125, max_mean=3, min_disp=0.5)
    elif not n_var_genes is None:
        sc.pp.highly_variable_genes(adata_deg_var, n_top_genes=n_var_genes)
    if var_genes is None:
        adata_deg = adata_deg[:, adata_deg_var.var.highly_variable]
    else:
        adata_deg = adata_deg[:, var_genes]
    del adata_deg_var

    summary_name = 'commot-'+database_name+'-sum-'+summary
    if summary == 'sender':
        summary_abrv = 's'
    else:
        summary_abrv = 'r'
    if not pathway_name is None:
        comm_sum = adata.obsm[summary_name][summary_abrv+'-'+pathway_name].values.reshape(-1,1)
    elif pathway_name is None:
        comm_sum = adata.obsm[summary_name][summary_abrv+'-'+lr_pair[0]+'-'+lr_pair[1]].values.reshape(-1,1)
    cell_weight = np.ones_like(comm_sum).reshape(-1,1)

    # send adata to R
    adata_r = anndata2ri.py2rpy(adata_deg)
    ro.r.assign("adata", adata_r)
    ro.r("X <- as.matrix( assay( adata, 'X') )")
    ro.r.assign("pseudoTime", comm_sum)
    ro.r.assign("cellWeight", cell_weight)

    # perform analysis (tradeSeq-1.0.1 in R-3.6.3)
    string_fitGAM = 'sce <- fitGAM(counts=X, pseudotime=pseudoTime[,1], cellWeights=cellWeight[,1], nknots=%d, verbose=TRUE)' % nknots
    ro.r(string_fitGAM)
    ro.r('assoRes <- data.frame( associationTest(sce, global=FALSE, lineage=TRUE) )')
    ro.r('assoRes[is.nan(assoRes[,"waldStat_1"]),"waldStat_1"] <- 0.0')
    ro.r('assoRes[is.nan(assoRes[,"df_1"]),"df_1"] <- 0.0')
    ro.r('assoRes[is.nan(assoRes[,"pvalue_1"]),"pvalue_1"] <- 1.0')
    with localconverter(ro.pandas2ri.converter):
        df_assoRes = ro.r['assoRes']
    ro.r('assoRes = assoRes[assoRes[,"pvalue_1"] <= %f,]' % deg_pvalue_cutoff)
    ro.r('oAsso <- order(assoRes[,"waldStat_1"], decreasing=TRUE)')
    if n_deg_genes is None:
        n_deg_genes = df_assoRes.shape[0]
    string_cluster = 'clusPat <- clusterExpressionPatterns(sce, nPoints = %d,' % n_points\
        + 'verbose=TRUE, genes = rownames(assoRes)[oAsso][1:min(%d,length(oAsso))],' % n_deg_genes \
        + ' k0s=4:5, alphas=c(0.1))'
    ro.r(string_cluster)
    ro.r('yhatScaled <- data.frame(clusPat$yhatScaled)')
    with localconverter(ro.pandas2ri.converter):
        yhat_scaled = ro.r['yhatScaled']

    df_deg = df_assoRes.rename(columns={'waldStat_1':'waldStat', 'df_1':'df', 'pvalue_1':'pvalue'})
    idx = np.argsort(-df_deg['waldStat'].values)
    df_deg = df_deg.iloc[idx]
    df_yhat = yhat_scaled

    anndata2ri.deactivate()
    ro.numpy2ri.deactivate()
    ro.pandas2ri.deactivate()

    return df_deg, df_yhat
    
def communication_deg_clustering(
    df_deg: pd.DataFrame,
    df_yhat: pd.DataFrame,
    deg_clustering_npc: int = 10,
    deg_clustering_knn: int = 5,
    deg_clustering_res: float = 1.0,
    n_deg_genes: int = 200,
    p_value_cutoff: float = 0.05
):
    """
    Cluster the communcation DE genes based on their fitted expression pattern.

    Parameters
    ----------
    df_deg
        The deg analysis summary data frame obtained by running ``commot.tl.communication_deg_detection``.
        Each row corresponds to one tested genes and columns include "waldStat" (Wald statistics), "df" (degrees of freedom), and "pvalue" (p-value of the Wald statistics).
    df_yhat
        The fitted (smoothed) gene expression pattern obtained by running ``commot.tl.communication_deg_detection``.
    deg_clustering_npc
        Number of PCs when performing PCA to cluster gene expression patterns
    deg_clustering_knn
        Number of neighbors when constructing the knn graph for leiden clustering.
    deg_clustering_res
        The resolution parameter for leiden clustering.
    n_deg_genes
        Number of top deg genes to cluster.
    p_value_cutoff
        The p-value cutoff for genes to be included in clustering analysis.

    Returns
    -------
    df_deg_clus: pd.DataFrame
        A data frame of clustered genes.
    df_yhat_clus: pd.DataFrame
        The fitted gene expression patterns of the clustered genes

    """
    df_deg = df_deg[df_deg['pvalue'] <= p_value_cutoff]
    n_deg_genes = min(n_deg_genes, df_deg.shape[0])
    idx = np.argsort(-df_deg['waldStat'])
    df_deg = df_deg.iloc[idx[:n_deg_genes]]
    yhat_scaled = df_yhat.loc[df_deg.index]
    x_pca = PCA(n_components=deg_clustering_npc, svd_solver='full').fit_transform(yhat_scaled.values)
    cluster_labels = leiden_clustering(x_pca, k=deg_clustering_knn, resolution=deg_clustering_res, input='embedding')

    data_tmp = np.concatenate((df_deg.values, cluster_labels.reshape(-1,1)),axis=1)
    df_metadata = pd.DataFrame(data=data_tmp, index=df_deg.index,
        columns=['waldStat','df','pvalue','cluster'] )
    return df_metadata, yhat_scaled

def communication_impact(
    adata: anndata.AnnData,
    database_name: str = None,
    pathway_name: str = None,
    pathway_sum_only: bool = False,
    heteromeric_delimiter: str = '_',
    normalize: bool = False,
    method: str = None,
    corr_method: str = "spearman",
    tree_method: str = "rf",
    tree_ntrees: int = 100,
    tree_repeat: int = 100,
    tree_max_depth: int = 5,
    tree_max_features: str = 'sqrt',
    tree_learning_rate: float = 0.1,
    tree_subsample: float = 1.0,
    tree_combined: bool = False,
    ds_genes: list = None,
    bg_genes: Union[list, int] = 100
):
    """
    Analyze impact of communication.

    When using the 'treebased_score' as the method, there is potentially dilution of importance between the LR pairs if 'tree_combined' is set to True.
    Therefore, if uniqueness of potential impact of various LR pairs on the target genes is not the focus, 'tree_combined' can be set to False.
    If the unique impact of signaling in addition to the intra-cellular regulatory impact of target genes is not of interest, 'bg_genes' can be set to 0.

    Parameters
    ----------
    adata
        The data matrix of shape ``n_obs`` × ``n_var``.
        Rows correspond to cells or positions and columns to genes.
        The full normalized dataset should be available in ``adata.raw``.
    database_name
        Name of the ligand-receptor database. 
    pathway_name
        Name of the signaling pathway.
    pathway_sum_only
        If ``True``, examine only the total signaling activity sum over signaling pathways without looking at individual ligand-receptor pairs.
    heteromeric_delimiter
        The delimiter that separates heteromeric units in the ligand-receptor database.
    normalize
        Whether to perform normalization before determining variable genes.
    method
        'partial_corr': partial correlation.
        'semipartial_corr': semipartial correlation.
        'treebased_score': machine learning based score (ensemble of trees).
    corr_method
        The correlation coefficient to use when method is 'partial_corr' or 'semipartial_corr'.
        'spearman': Spearman's r. 'pearson': Pearson's r.
    tree_method
        The ensemble of trees method to use when method is 'treebased_score'.
        'gbt': gradient boosted trees. 'rf': random forest.
    tree_ntrees
        Number of trees when using 'treebased_score'.
    tree_repeat
        Number of times to repeat to account for randomness when using 'treebased_score'.
    tree_mas_depth
        Max depth of trees when using 'treebased_score'.
    tree_max_features
        Max features for trees when using 'treebased_score'.
    tree_learning_rate
        Learning rate when using 'treebased_score'.
    tree_subsample
        Subsample (between 0 and 1) when using 'treebased_score'.
    tree_combined
        If True, use a single model for each target gene with all features.
    ds_genes
        A list of genes for analyzing the correlation with cell-cell communication, for example, the highly variable genes.
    bg_genes
        If an integer, the top number of variable genes are used.
        Alternatively, a list of genes.

    Returns
    -------
    df_impact: pd.DataFrame
        A data frame describing the correlation 
        between the ds_genes and cell-cell communication.
    """

    # Get a list of background genes using most 
    # variable genes if only given a number.
    adata_bg = adata.raw.to_adata()
    adata_all = adata.raw.to_adata()
    if normalize:
        sc.pp.normalize_total(adata_bg, inplace=True)
        sc.pp.log1p(adata_bg)
    if np.isscalar(bg_genes):
        ng_bg = int(bg_genes)
        sc.pp.highly_variable_genes(adata_bg, n_top_genes=ng_bg)
        adata_bg = adata_bg[:,adata_bg.var.highly_variable]
    else:
        adata_bg = adata_bg[:,bg_genes]
    # Prepare downstream or upstream genes
    ncell = adata.shape[0]
    col_names = []
    Ds_exps = []
    Ds_exp_total = np.zeros([ncell], float)
    for i in range(len(ds_genes)):
        Ds_exp = np.array(adata_all[:,ds_genes[i]].X.toarray()).reshape(-1)
        Ds_exps.append(Ds_exp)
        col_names.append(ds_genes[i])
        Ds_exp_total += Ds_exp
    Ds_exps.append(Ds_exp_total); col_names.append('average')
    # Impact analysis
    df_ligrec = adata.uns['commot-'+database_name+'-info']['df_ligrec']
    available_pathways = []
    for i in range(df_ligrec.shape[0]):
        _, _, tmp_pathway = df_ligrec.iloc[i,:]
        if not tmp_pathway in available_pathways:
            available_pathways.append(tmp_pathway)
    pathway_genes = [[] for i in range(len(available_pathways))]
    all_lr_genes = []
    for i in range(df_ligrec.shape[0]):
        tmp_lig, tmp_rec, tmp_pathway = df_ligrec.iloc[i,:]
        idx = available_pathways.index(tmp_pathway)
        tmp_ligs = tmp_lig.split(heteromeric_delimiter)
        tmp_recs = tmp_rec.split(heteromeric_delimiter)
        for lig in tmp_ligs:
            if not lig in pathway_genes[idx]:
                pathway_genes[idx].append(lig)
            if not lig in all_lr_genes:
                all_lr_genes.append(lig)
        for rec in tmp_recs:
            if not rec in pathway_genes[idx]:
                pathway_genes[idx].append(rec)
            if not rec in all_lr_genes:
                all_lr_genes.append(rec)
    bg_genes = list( adata_bg.var_names )

    sum_names = []
    exclude_lr_genes_list = []
    if pathway_name is None and not pathway_sum_only:
        for i in range(df_ligrec.shape[0]):
            tmp_lig, tmp_rec, _ = df_ligrec.iloc[i,:]
            sum_names.append("%s-%s" % (tmp_lig, tmp_rec))
            exclude_lr_genes_list.append(set(tmp_lig.split(heteromeric_delimiter)).union(set(tmp_rec.split(heteromeric_delimiter))))
        for tmp_pathway in available_pathways:
            sum_names.append(tmp_pathway)
            exclude_lr_genes_list.append(set(pathway_genes[available_pathways.index(tmp_pathway)]))
        sum_names.append('total-total')
        exclude_lr_genes_list.append(set(all_lr_genes))
    elif not pathway_name is None and not pathway_sum_only:
        for i in range(df_ligrec.shape[0]):
            tmp_lig, tmp_rec, tmp_pathway = df_ligrec.iloc[i,:]
            if tmp_pathway == pathway_name:
                sum_names.append("%s-%s" % (tmp_lig, tmp_rec))
                exclude_lr_genes_list.append(set(tmp_lig.split(heteromeric_delimiter)).union(set(tmp_rec.split(heteromeric_delimiter))))
        sum_names.append(pathway_name)
        exclude_lr_genes_list.append(set(pathway_genes[available_pathways.index(pathway_name)]))

    elif pathway_sum_only:
        sum_names = available_pathways
        for i in range(len(available_pathways)):
            exclude_lr_genes_list.append(set(pathway_genes[i]))

    nrows = 2 * len(sum_names)

    ncols = len(ds_genes) + 1
    impact_mat = np.empty([nrows, ncols], float)
    
    row_names_sender = []; row_names_receiver = []
    exclude_lr_genes_list = []
    for i in range(len(sum_names)):
        row_names_sender.append('s-%s' % sum_names[i])
        row_names_receiver.append('r-%s' % sum_names[i])
    row_names = row_names_sender + row_names_receiver
    exclude_lr_genes_list = exclude_lr_genes_list + exclude_lr_genes_list

    print(nrows, ncols)
    for j in range(ncols):
        print(j)
        if j == ncols-1:
            exclude_ds_genes = set(ds_genes)
        else:
            exclude_ds_genes = set([ds_genes[j]])
        if method == 'treebased_score' and tree_combined:
            exclude_lr_genes = set(all_lr_genes)
            exclude_genes = list(exclude_lr_genes.union(exclude_ds_genes))
            use_genes = list( set(bg_genes) - set(exclude_genes) )
            bg_mat = np.array( adata_bg[:,use_genes].X.toarray() )
            sum_mat = np.concatenate((adata.obsm['commot-'+database_name+'-sum-sender'][row_names_sender].values, \
                adata.obsm['commot-'+database_name+'-sum-receiver'][row_names_receiver].values), axis=1)
            r = treebased_score_multifeature(sum_mat, Ds_exps[j], bg_mat,
                n_trees = tree_ntrees, n_repeat = tree_repeat,
                max_depth = tree_max_depth, max_features = tree_max_features,
                learning_rate = tree_learning_rate, subsample = tree_subsample)
            impact_mat[:,j] = r[:]
        else:
            for i in range(nrows):
                row_name = row_names[i]
                exclude_lr_genes = exclude_lr_genes_list[i]

                exclude_genes = list(exclude_lr_genes.union(exclude_ds_genes))
                use_genes = list( set(bg_genes) - set(exclude_genes) )
                bg_mat = np.array( adata_bg[:,use_genes].X.toarray() )
                if row_name[0] == 's':
                    sum_vec = adata.obsm['commot-'+database_name+'-sum-sender'][row_name].values.reshape(-1,1)
                elif row_name[0] == 'r':
                    sum_vec = adata.obsm['commot-'+database_name+'-sum-receiver'][row_name].values.reshape(-1,1)
                if method == "partial_corr":
                    r,p = partial_corr(sum_vec, Ds_exps[j].reshape(-1,1), bg_mat, method=corr_method)
                elif method == "semipartial_corr":
                    r,p = semipartial_corr(sum_vec, Ds_exps[j].reshape(-1,1), ycov=bg_mat, method=corr_method)
                elif method == "treebased_score":
                    r = treebased_score(sum_vec, Ds_exps[j], bg_mat,
                        n_trees = tree_ntrees, n_repeat = tree_repeat,
                        max_depth = tree_max_depth, max_features = tree_max_features,
                        learning_rate = tree_learning_rate, subsample = tree_subsample)
                impact_mat[i,j] = r
    df_impact = pd.DataFrame(data=impact_mat, index = row_names, columns = col_names)
    return df_impact


# Has not adapted new naming scheme
def group_cluster_communication(
    adata: anndata.AnnData,
    clustering: str = None,
    cluster_permutation_type: str = 'label',
    keys = None,
    p_value_cutoff: float = 0.05,
    quantile_cutoff: float = 0.99,
    dissimilarity_method: str = None,
    leiden_k: int = 5,
    leiden_resolution: float = 1.0,
    leiden_random_seed: int = 1,
    leiden_n_iterations: int = -1,
    d_global_structure_weights: tuple = (0.45,0.45,0.1)
):
    """
    Idenfitify groups of cluster-cluster communication with similar
    pattern.

    The CCC inference should have been summarized to cluster level using either :func:`commot.tl.cluster_communication` function 
    or :func:`commot.tl.cluster_communication_spatial_permutation` function. The dissimilarities among different ligand-receptor pairs or signaling pathways
    are first quantified which will be used with the leiden clustering algorithm [Traag2019]_ to cluster these CCC networks.

    Parameters
    ----------
    adata
        The data matrix with the cluster-cluster communication 
        info stored in ``adata.uns``.
    clustering
        Name of clustering with the labels stored in ``.obs[clustering]``.
    cluster_permutation_type
        ``'label'`` if the function :func:`commot.tl.cluster_communication` was used and ``'spatial'`` if :func:`commot.tl.cluster_communication_spatial_permutation` was used.
    keys
        A list of keys for the analyzed communication connections as strings.
        For example, the string ``'databaseX-pathwayX'`` represents the cluster-level CCC of signaling pathway "pathwayX" computed with the LR database "databaseX".
        For another example, the string ``'databaseX-ligA-recA'`` represents the cluster-level CCC of the LR pair "ligA-recA" computed with the LR database "databaseX". 
    p_value_cutoff
        The cutoff of p-value for including an edge.
    quantile_cutoff
        The quantile cutoff for including an edge. Set to 1 to disable this criterion.
        The quantile_cutoff and p_value_cutoff works in the "or" logic to avoid missing
        significant signaling connections. 
        An edge will be ignored only if it has a p-value greater than the ``p_value_cutoff`` and a score smaller than the score quantile cutoff.
    dissimilarity_method
        The dissimilarity measurement between graphs to use. 
        'jaccard' for Jaccard distance.
        'jaccard_weighted' for weighted Jaccard distance.
        'global_structure' for a metric focusing on global structure [Schieber2017]_.
    leiden_k
        Number of neighbors for the knn-graph for using leiden clustering algorithm.
    leiden_resolution
        The resolution parameter for the leiden clustering algorithm.
    leiden_random_seed
        The random seed for the leiden clustering algorithm.
    leiden_n_iterations
        The maximum number of iterations for the leiden algorithm.
        The algorithm will run until convergence if set to -1.
    d_global_structure_weights
        The weights for the three terms in the global structural dissimilarity.
        See [Schieber2017]_ for more information.
    
    Returns
    -------
    communication_clusterid : np.ndarray
        The group id of the cluster-cluster communications.
    D : np.ndarray
        The dissimilarity matrix for the cluster-cluster communications.

    References
    ----------
    .. [Schieber2017] Schieber, T. A., Carpi, L., Díaz-Guilera, A., Pardalos, 
        P. M., Masoller, C., & Ravetti, M. G. (2017). Quantification 
        of network structural dissimilarities. Nature communications, 8(1), 1-10.
    .. [Traag2019] Traag, V. A., Waltman, L., & Van Eck, N. J. (2019). 
        From Louvain to Leiden: guaranteeing well-connected communities. Scientific reports, 9(1), 1-12.

    """
    
    # Get a list of filtered communication matrices corresponding to the keys.
    As = []
    for key in keys:
        if cluster_permutation_type == 'label':
            tmp_name = 'commot_cluster-%s-%s' % (clustering,key)
        elif cluster_permutation_type == 'spatial':
            tmp_name = 'commot_cluster_spatial_permutation-%s-%s' % (clustering,key)
        X_tmp = adata.uns[tmp_name]['communication_matrix'].values.copy()
        pvalue_tmp = adata.uns[tmp_name]['communication_pvalue'].values.copy()
        if not quantile_cutoff is None:
            cutoff = np.quantile(X_tmp.reshape(-1), quantile_cutoff)
        else:
            cutoff = np.inf
        tmp_mask = ( X_tmp < cutoff ) * ( pvalue_tmp > p_value_cutoff )
        X_tmp[tmp_mask] = 0
        As.append(X_tmp)
    # Get a distance/dissimilarity matrix between the communication matrices.
    D = np.zeros([len(keys), len(keys)], float)
    for i in range(len(keys)-1):
        for j in range(i+1,len(keys)):
            if dissimilarity_method == 'jaccard':
                d = d_graph_local_jaccard(As[i], As[j])
            elif dissimilarity_method == 'jaccard_weighted':
                d = d_graph_local_jaccard_weighted(As[i], As[j])
            elif dissimilarity_method == 'global_structure':
                w1,w2,w3 = d_global_structure_weights
                d = d_graph_global_structure(As[i], As[j], w1=w1, w2=w2, w3=w3)
            D[i,j] = d; D[j,i] = d
    # Perform clustering
    leiden_k = min(leiden_k, len(keys)-1)
    communication_clusterid = leiden_clustering(D,
        k = leiden_k, resolution = leiden_resolution,
        random_seed = leiden_random_seed, 
        n_iterations = leiden_n_iterations)
    
    return communication_clusterid, D

def group_cell_communication(
    adata: anndata.AnnData,
    keys = None,
    bin_method: str = 'gaussian_mixture',
    bin_append_zeros: str = 'full',
    bin_random_state: int = 1,
    bin_cutoff: float = 0,
    knn: int = 2,
    dissimilarity_method: str = 'graphwave',
    kw_graphwave: dict = {'sample_number':200, 'step_size':0.1, 'heat_coefficient': 1.0, 'approximation':100, 'mechanism':'approximation', 'switch':1000, 'seed':42},
    leiden_k: int = 5,
    leiden_resolution: float = 1.0,
    leiden_random_seed: int = 1,
    leiden_n_iterations: int = -1
):
    """
    Idenfitify groups of cell-cell communication with similar
    pattern.

    The cell-cell communication should have been computed by the function :func:`commot.tl.spatial_communication`.

    Parameters
    ----------
    adata
        The data matrix with the cell-cell communication 
        info stored in ``adata.obsp``.
    keys
        A list of keys for the analyzed communication connections as strings.
        For example, the string ``'databaseX-pathwayX'`` represents the CCC of signaling pathway "pathwayX" computed with the LR database "databaseX".
        For another example, the string ``'databaseX-ligA-recA'`` represents the CCC of the LR pair "ligA-recA" computed with the LR database "databaseX".
        The cell-level CCC networks corresponding to the above examples are stored in ``.obsp['commot-databaseX-pathwayX]`` and ``.obsp['commot-databaseX-ligA-recA']``.
    bin_method
        Method for binarize communication connections. Choices: 'gaussian_mixture', 'kmeans'.
    bin_append_zeros
        Number of zeros to append to the non-zero entries when running the binarization.
        'full' use the full flattened cell-by-cell communication matrix.
        'match' append the same number of zeros to the vector of non-zero entries.
    bin_random_state
        The random seed for binarization methods.
    bin_cutoff
        Force all connections with a weight smaller than or equal to bin_cutoff to be zero
        , regardless of binarization result.
    knn
        Number of neighbors for building the spatial knn graph.
    dissimilarity_method
        The method for quantifying dissimilarity.
        'graphwave', node structural embedding by GraphWave [Donnat2018]_ implemented 
        in `Karate Club
        <https://github.com/benedekrozemberczki/karateclub/>`_.
    kw_graphwave
        Keywords for GraphWave. Defaults: {'sample_number':200, 'step_size':0.1, 'heat_coefficient': 1.0,
        'approximation':100, 'mechanism':'approximation', 'switch':1000, 'seed':42} See details at `Karate Club
        <https://github.com/benedekrozemberczki/karateclub/>`_.
    leiden_k
        Number of neighbors for the knn-graph to be fed to leiden clustering algorithm.
    leiden_resolution
        The resolution parameter for the leiden clustering algorithm.
    leiden_random_seed
        The random seed for the leiden clustering algorithm.
    leiden_n_iterations
        The maximum number of iterations for the leiden algorithm.
        The algorithm will run until convergence if set to -1.
    
    Returns
    -------
    communication_clusterid : np.ndarray
        The group id of the cell-cell communications.
    D : np.ndarray
        The dissimilarity matrix for the cell-cell communications.
    

    References
    ----------
    .. [Donnat2018] Donnat, C., Zitnik, M., Hallac, D., & Leskovec, J. (2018, July). Learning structural node embeddings 
        via diffusion wavelets. In Proceedings of the 24th ACM SIGKDD International 
        Conference on Knowledge Discovery & Data Mining (pp. 1320-1329).

    """
    
    nkey = len(keys)
    ncell = adata.shape[0]
    import karateclub
    # Get a dissimilarity matrix D
    if dissimilarity_method == 'graphwave':
        if knn > 0:
            A_knn = kneighbors_graph(adata.obsm['spatial'],
                knn, mode = 'connectivity', include_self = False)
            A_spatial = A_knn + A_knn.T
            A_spatial.eliminate_zeros()
            A_spatial.data[:] = 1
        elif knn == 0:
            A_spatial = sparse.csr_matrix((ncell,ncell))
        heat_mats = []
        for key in keys:
            A_signal = adata.obsp['commot-%s' % key]
            A_signal_bin = binarize_sparse_matrix(A_signal, method = bin_method,
                append_zeros = bin_append_zeros, random_state = bin_random_state)
            A_signal_bin_sym = A_signal_bin + A_signal_bin.T
            A_signal_bin_sym.eliminate_zeros()
            A_signal_bin_sym.data[:] = 1
            A = A_spatial + A_signal_bin_sym
            gw = karateclub.GraphWave(**kw_graphwave)
            G = nx.from_scipy_sparse_matrix(A)
            gw.fit(G)
            R = gw.get_embedding()
            heat_mats.append(R)
        D = np.zeros([nkey,nkey],float)
        for i in range(nkey-1):
            for j in range(i+1, nkey):
                d = np.sqrt( np.linalg.norm(heat_mats[i] - heat_mats[j]) ** 2 / float(ncell) )
                D[i,j] = d; D[j,i] = d

    # Perform clustering on D
    leiden_k = min(leiden_k, len(keys)-1)
    communication_clusterid = leiden_clustering(D,
        k = leiden_k, resolution = leiden_resolution,
        random_seed = leiden_random_seed, 
        n_iterations = leiden_n_iterations)

    return communication_clusterid, D


def group_communication_direction(
    adata: anndata.AnnData,
    keys: list = None,
    summary: str = 'sender',
    knn_smoothing: int = -1,
    normalize_vf: str = 'quantile',
    normalize_quantile: float = 0.99,
    dissimilarity_method: str = 'dot_product',
    leiden_k: int = 5,
    leiden_resolution: float = 1.0,
    leiden_random_seed: int = 1,
    leiden_n_iterations: int = -1
):
    """
    Idenfitify groups of communication directions with similar
    pattern.

    The cell-cell communication should have been computed by the function :func:`commot.tl.spatial_communication`.
    The cell-cell communication direction should have been computed by the function :func:`commot.tl.communication_direction`.

    Parameters
    ----------
    adata
        The data matrix with the communication direction
        info stored in ``adata.obsm``, e.g., ``.obsm['commot_sender_vf-databaseX-ligA-recA']`` stores the CCC direction of the LR pair 'ligA-recA' computed
        with the LR database 'databaseX' summarized in the signal senders' perspective 
    keys
        A list of keys for the analyzed communication connections as strings.
        For example, the string ``'databaseX-pathwayX'`` represents the CCC of signaling pathway "pathwayX" computed with the LR database "databaseX".
        For another example, the string ``'databaseX-ligA-recA'`` represents the CCC of the LR pair "ligA-recA" computed with the LR database "databaseX".
        The computed CCC direction corresponding to the above examples (summarized as 'sent to' or 'received from' directions) should be available in 
        ``.obsm['commot_sender_vf-databaseX-pathwayX]``,  ``.obsm['commot_receiver_vf-databaseX-pathwayX]``and
        ``.obsm['commot_sender_vf-databaseX-ligA-recA']``, ``.obsm['commot_receiver_vf-databaseX-ligA-recA']``.
    summary
        If 'sender', use the vector field describing to which direction the signals are sent to.
        If 'receiver', use the vector field describing from which direction the signals are received from.
    knn_smoothing
        The number of neighbors to smooth the communication direction. 
        If -1, no smoothing is performed.
    normalize_vf
        If 'quantile', divide all values by the length 
        given by the normalize_quantile parameter.
        If 'unit_norm', normalize each individual vector into unit norm.
        If None, original unit is used.
    normalize_quantile
        The quantile parameter to use if normalize_vf is set to 'quantile'.
    dissimilarity_method
        Currently, only dot_product is implemented.
    leiden_k
        Number of neighbors for the knn-graph to be fed to leiden clustering algorithm.
    leiden_resolution
        The resolution parameter for the leiden clustering algorithm.
    leiden_random_seed
        The random seed for the leiden clustering algorithm.
    leiden_n_iterations
        The maximum number of iterations for the leiden algorithm.
        The algorithm will run until convergence if set to -1.

    Returns
    -------
    direction_clusterid : np.ndarray
        The group id of the communication directions.
    D : np.ndarray
        The dissimilarity matrix for the communication directions.
    
    """
    
    # Process the vector fields
    V_list = []
    for key in keys:
        V = adata.obsm['commot_%s_vf-%s' % (summary, key)]
        V_processed = preprocess_vector_field(adata.obsm['spatial'],
            V, knn_smoothing = knn_smoothing, normalize_vf = normalize_vf,
            quantile = normalize_quantile)
        V_list.append(V_processed)
    # Get a distance matrix between the vector fields
    nV = len(keys)
    D = np.zeros([nV,nV], float)
    for i in range(nV-1):
        Vi = V_list[i]
        for j in range(i+1,nV):
            Vj = V_list[j]
            if dissimilarity_method == 'dot_product':
                d = np.exp( - ( ( Vi * Vj ).sum(axis=1) ).mean() )
            D[i,j] = d; D[j,i] = d
    # Cluster the vector fields with the D matrix
    # Perform clustering
    leiden_k = min(leiden_k, len(keys)-1)
    direction_clusterid = leiden_clustering(D,
        k = leiden_k, resolution = leiden_resolution,
        random_seed = leiden_random_seed, 
        n_iterations = leiden_n_iterations)
    
    return direction_clusterid, D


def communication_spatial_autocorrelation(
    adata: anndata.AnnData,
    keys: list = None,
    method: str = 'Moran',
    normalize_vf: bool = False,
    summary: str = 'sender',
    weight_bandwidth: float = None,
    weight_k: int = 10,
    weight_function: str = 'triangular',
    weight_row_standardize: bool = False,
    n_permutations: int = 999
):
    """
    Spatial autocorrelation of communication directions.

    The spatial autocorrelation helps to detect spatial regions within which the CCC directions are similar.
    The cell-cell communication should have been computed by the function :func:`commot.tl.spatial_communication`.
    The cell-cell communication direction should have been computed by the function :func:`commot.tl.communication_direction`.

    Parameters
    ----------
    adata
        The data matrix with the communication vector fields
        info stored in ``adata.ubsm``.
    keys
        A list of keys for the analyzed communication connections as strings.
        For example, the string ``'databaseX-pathwayX'`` represents the CCC of signaling pathway "pathwayX" computed with the LR database "databaseX".
        For another example, the string ``'databaseX-ligA-recA'`` represents the CCC of the LR pair "ligA-recA" computed with the LR database "databaseX".
        The computed CCC direction corresponding to the above examples (summarized as 'sent to' or 'received from' directions) should be available in 
        ``.obsm['commot_sender_vf-databaseX-pathwayX]``,  ``.obsm['commot_receiver_vf-databaseX-pathwayX]``and
        ``.obsm['commot_sender_vf-databaseX-ligA-recA']``, ``.obsm['commot_receiver_vf-databaseX-ligA-recA']``.
    method
        The method to use. Currently, only Moran's I [Liu2015]_ for vectors is implemented.
    normalize_vf
        Whether to normalize the vector field so that the autocorrelation only reflects
        directions.
    summary
        If 'sender', use the vector field describing to which direction the signals are sent.
        If 'receiver', use the vector field describing from which direction the signals are from.
    weight_bandwidth
        The bandwidth for the kernel to assign knn graph weights.
        If given, weight_k is ignored.
    weight_k
        The number of nearest neighbors for the knn graph.
    weight_function
        Kernel functions for assigning weight. 
        Choices: 'triangular','uniform','quadratic','quartic','gaussian'.
        See libpysal.weights.Kernel of the ``libpysal`` package for details.
    weight_row_standardize
        Whether to standardize the weights so that the heterogeneity in local cell/position
        density does not affect the results.
    n_permutations
        Number of permutations for computing p-values.

    Returns
    -------
    moranI : np.ndarray
        A vector of moran's I statistics for corresponding to each key in keys.
    p_value : np.ndarray
        The p-values.

    References
    ----------
    .. [Liu2015] Liu, Y., Tong, D., & Liu, X. (2015). Measuring spatial 
        autocorrelation of vectors. Geographical Analysis, 47(3), 300-319.

    """
    
    moranI = []
    p_value = []
    X = adata.obsm['spatial']
    for key in keys:
        V = adata.obsm['commot_%s_vf-%s' % (summary, key)]
        if normalize_vf:
            V = normalize(V)
        I,p = moranI_vector_global(X, V,
            weight_bandwidth = weight_bandwidth,
            weight_k = weight_k,
            weight_function = weight_function,
            weight_row_standardize = weight_row_standardize,
            n_permutations = n_permutations)
        moranI.append(I)
        p_value.append(p)
    moranI = np.array(moranI, float)
    p_value = np.array(p_value, float)

    return moranI, p_value





import os

import numpy as np
import pandas as pd
import scanpy as sc
import matplotlib.pyplot as plt
import seaborn as sns

from sklearn.metrics import r2_score
from scipy.stats import pearsonr, gaussian_kde


# Module import
from .post_analysis import get_z_umap
from .utils import extract_feature


def plot_integrated_spatial_feature(data, 
                                    sample_id,
                                    vmin=None,
                                    vmax=None,
                                    spot_size = 1,
                                    figsize=(2.5, 2),
                                    dpi = 80,
                                    cmap = 'Blues',
                                    legend_on = True,
                                    legend_loc = (2,0.5)
                                   ):

    adata_ann = data[data.obs['sample']==sample_id]
    #color_idx_list = np.array(adata_ann.obs['pheno_louvain']).astype(int)
    color_idx_list = np.array(adata_ann.obs['pheno_louvain']).astype(int)
    all_loc = np.array(adata_ann.obsm['spatial'])
    fig,axs= plt.subplots(1,1,figsize=figsize,dpi=dpi)
    for i in np.unique(color_idx_list):
        g=axs.scatter(all_loc[color_idx_list==i,0],-all_loc[color_idx_list==i, 1],s=spot_size,marker='h',c=cmap[i])

    if legend_on:
        plt.legend(list(np.unique(color_idx_list)),title='Hub region',loc='right',bbox_to_anchor=legend_loc)
    #fig.colorbar(g,label=label)
    #plt.title(sample_id)
    axs.set_xticks([])
    axs.set_yticks([])
    plt.axis('off')
    return axs
    
def plot_spatial_density(
    data,
    vmin=None,
    vmax=None,
    spot_size = 1,
    figsize=(2.5, 2),
    dpi = 80,
    cmap = 'Blues',
    colorbar_on = True,
):
    fig, axes = plt.subplots(1, 1, figsize=figsize, dpi=dpi)
    g = axes.scatter(data.obsm['spatial'][:,0],
                     -data.obsm['spatial'][:,1],
                     c=data.obsm['ql_m'],
                     vmin=vmin, 
                     vmax=vmax,
                     cmap=cmap,
                     s=spot_size,
                    )
    if colorbar_on:
        fig.colorbar(g, label='Estimated density')
    axes.axis('off')
    return axes

def plot_spatial_cell_type_frac( data = None,
                                 cell_type = None,
                                 vmin=None,# adjust 
                                 vmax=None,# adjust 
                                 spot_size=2,# adjust 
                                 figsize = (3,2.5),
                                 fig_dpi = 80, # >300 for high quality img
                                 cmap = 'magma', 
                                 colorbar_on = True,
                                 label=None,
                                 title=None,
                                ):
    
    
    fig,axs= plt.subplots(1,1,figsize=figsize,dpi=fig_dpi)
    idx = data.uns['cell_types'].index(cell_type)
    g=axs.scatter(data.obsm['spatial'][:,0],
                  -data.obsm['spatial'][:,1],
                  c=data.obsm['qc_m'][:,idx],
                  cmap=cmap,
                  vmin=vmin,
                  vmax=vmax,
                  s=spot_size
                 )
    if title is not None:
        plt.title(title)
    else:
        plt.title(cell_type)
    if colorbar_on:
        fig.colorbar(g,label=label)
    plt.axis('off')
    return ax
    
    
    
def plot_z_umap_cell_type_frac(data = None,
                             cell_type = None,
                             vmin=None,# adjust 
                             vmax=None,# adjust 
                             spot_size=2,# adjust 
                             figsize = (3,2.5),
                             dpi = 80, # >300 for high quality img
                             cmap = 'magma', 
                             colorbar_on = True,
                             label=None,
                             title=None,
                        ):
    
    
    fig,axs= plt.subplots(1,1,figsize=figsize,dpi=dpi)
    idx = data.uns['cell_types'].index(cell_type)
    g=axs.scatter(data.obsm['z_umap'][:,0],
                  data.obsm['z_umap'][:,1],
                  c=data.obsm['qc_m'][:,idx],
                  cmap=cmap,
                  vmin=vmin,
                  vmax=vmax,
                  s=spot_size
                 )
    if title is not None:
        plt.title(title)
    else:
        plt.title(cell_type)
    if colorbar_on:
        fig.colorbar(g,label=label)
    plt.axis('off')
    return axs
        
def plot_spatial_feature(data = None,
                         feature = None,
                         vmin=None,# adjust 
                         vmax=None,# adjust 
                         spot_size=2,# adjust 
                         figsize = (3,2.5),
                         dpi = 80, # >300 for high quality img
                         cmap = 'magma', 
                         colorbar_on = True,
                         label=None
                        ):
    
    
    fig,axs= plt.subplots(1,1,figsize=figsize,dpi=dpi)

    g=axs.scatter(data.obsm['spatial'][:,0],
                  -data.obsm['spatial'][:,1],
                  c=feature,
                  cmap=cmap,
                  vmin=vmin,
                  vmax=vmax,
                  s=spot_size
                 )
    
    if colorbar_on:
        fig.colorbar(g,label=label)
    plt.axis('off')
    return axs

def plot_spatial_gene(data,
                      data_normed, 
                      gene_name = None,
                      log_gene = False,
                      vmin=None,
                      vmax=None,
                      spot_size=5,
                      figsize = (2,2),
                      dpi = 80,
                      cmap = 'magma',
                      colorbar_on = True,     
                     ):
    
    fig,axs= plt.subplots(1,1,figsize=figsize,dpi=fig_dpi)
    if log_gene:
        g=axs.scatter(data_normed.obsm['spatial'][:,0],
                      -data_normed.obsm['spatial'][:,1],
                      c=data_normed.to_df().loc[:,gene_name],
                      cmap=cmap,
                      vmin=vmin,
                      vmax=vmax,
                      s=spot_size
                     )
    else:
        g=axs.scatter(data.obsm['spatial'][:,0],
                      -data.obsm['spatial'][:,1],
                      c=data.to_df().loc[:,gene_name],
                      cmap=cmap,
                      vmin=vmin,
                      vmax=vmax,
                      s=spot_size
                     )
    if colorbar_on:
        fig.colorbar(g,label=gene_name)
    plt.axis('off')
    return axs


def plot_anchor_spots(umap_plot,
                      pure_spots,
                      sig_mean,
                      bbox_x=2,
                      dpi=80,
                      figsize=(3,3),
                      pure_spot_size=8,
                        all_spot_size=2):
                     
    fig,ax = plt.subplots(1,1,dpi=dpi,figsize=figsize)
    ax.scatter(umap_plot['umap1'],
               umap_plot['umap2'],
               s=all_spot_size,
               alpha=1,
               color='lightgray')
    for i in range(len(pure_spots)):
        ax.scatter(umap_plot['umap1'][pure_spots[i]],
                   umap_plot['umap2'][pure_spots[i]],
                   s=pure_spot_size)
    plt.legend(['all']+[i for i in sig_mean.columns],
               loc='right', 
               bbox_to_anchor=(bbox_x,0.5),)
    ax.grid(False)
    ax.axis('off')
    return ax


def plot_evs(evs, kmin, figsize=(6, 3), dpi=80):
    fig, ax = plt.subplots(1, 1, dpi=dpi, figsize=figsize)
    plt.plot(np.arange(len(evs))+kmin, evs, '.-')
    plt.xlabel('ks')
    plt.ylabel('Explained Variance')
    return ax


def pl_spatial_inf_feature(
    adata,
    feature,
    factor=None,
    vmin=0,
    vmax=None,
    spot_size=100,
    alpha=0,
    cmap='Spectral_r'
):
    """Spatial visualization of Starfysh inference features"""
    if isinstance(factor, str):
        assert factor in adata.uns['cell_types'], \
            "Invalid Starfysh inference factor (cell type): ".format(factor)
    elif isinstance(factor, list):
        for f in factor:
            assert f in adata.uns['cell_types'], \
                "Invalid Starfysh inference factor (cell type): ".format(f)
    else:
        factor = adata.uns['cell_types']  # if None, display for all cell types

    adata_pl = extract_feature(adata, feature)

    if feature == 'qc_m':
        if isinstance(factor, list):
            title = [f + ' (Inferred proportion - Spatial)' for f in factor]
        else:
            title = factor + ' (Inferred proportion - Spatial)'
        sc.pl.spatial(
            adata_pl,
            color=factor, spot_size=spot_size, color_map=cmap,
            ncols=3, vmin=vmin, vmax=vmax, alpha_img=alpha,
            title=title, legend_fontsize=8
        )
    elif feature == 'ql_m':
        title = 'Estimated tissue density'
        sc.pl.spatial(
            adata_pl,
            color='density', spot_size=spot_size, color_map=cmap,
            vmin=vmin, vmax=vmax, alpha_img=alpha,
            title=title, legend_fontsize=8
        )
    elif feature == 'qz_m':
        # Visualize deconvolution on UMAP of inferred Z-space
        qz_u = get_z_umap(adata_pl.obs.values)
        qc_df = extract_feature(adata, 'qc_m').obs
        if isinstance(factor, list):
            for cell_type in factor:
                title = cell_type + ' (Inferred proportion - UMAP of Z)'
                pl_umap_feature(qz_u, qc_df[cell_type].values, cmap, title,
                                vmin=vmin, vmax=vmax)
        else:
            title = factor + ' (Inferred proportion - UMAP of Z)'
            fig, ax = pl_umap_feature(qz_u, qc_df[factor].values, cmap, title,
                            vmin=vmin, vmax=vmax)
            return fig, ax
    else:
        raise ValueError('Invalid Starfysh inference results `{}`, please choose from `qc_m`, `qz_m` & `ql_m`'.format(feature))

    pass


def pl_umap_feature(qz_u, qc, cmap, title, spot_size=3, vmin=0, vmax=None,
                    dpi=80, figsize=(4, 3)):
    """Single Z-UMAP visualization of Starfysh deconvolutions"""
    fig, ax = plt.subplots(1, 1, dpi=dpi, figsize=figsize)
    g = ax.scatter(
        qz_u[:, 0], qz_u[:, 1],
        cmap=cmap, c=qc, s=spot_size, vmin=vmin, vmax=vmax,
    )
    ax.set_xticks([])
    ax.set_yticks([])
    ax.set_title(title)
    ax.axis('off')
    fig.colorbar(g, label='Inferred proportions')

    return fig, ax


def pl_spatial_inf_gene(
    adata=None,
    factor=None,
    feature=None,
    vmin=None,
    vmax=None,
    spot_size=100,
    alpha=0,
    figsize = (3,2.5),
    dpi = 80,
    cmap='Spectral_r',
    colorbar_on = True,          
    title = None ,          
    
):
    
    if isinstance(feature, str):
        assert feature in set(adata.var_names), \
            "Gene {0} isn't HVG, please choose from `adata.var_names`".format(feature)
        title_new = feature + ' (Predicted expression)'
    else:
        for f in feature:
            assert f in set(adata.var_names), \
                "Gene {0} isn't HVG, please choose from `adata.var_names`".format(f)
        title_new = [f + ' (Predicted expression)' for f in feature]

    if title is not None:
        title_new = title
    # Assign dummy `var_names` to avoid gene name in both obs & var
    adata_expr = extract_feature(adata, factor+'_inferred_exprs')
    adata_expr.var_names = np.arange(adata_expr.shape[1])
    
    #sc.settings.set_figure_params(figsize=figsize,dpi=fig_dpi,facecolor="white")
    
    sc.pl.spatial(
        adata_expr,
        color=feature, spot_size=spot_size, color_map=cmap,
        ncols=3, vmin=vmin, vmax=vmax, alpha_img=alpha,
        title=title_new,
        legend_fontsize=8,
        
    )
    
    pass


# --------------------------
# util funcs for benchmark
# --------------------------

def _dist2gt(A, A_gt):
    """
    Calculate the distance to ground-truth correlation matrix (proportions)
    """
    return np.linalg.norm(A - A_gt, ord='fro')


def _calc_rmse(y_true, y_pred):
    """Calculate per-spot RMSE between ground-truth & predicted proportions"""
    assert y_true.shape == y_pred.shape, "proportion matrices need to be the same shape to calculate RMSE"   
    n_cts = y_true.shape[1]
    rmse = np.sqrt(((y_true.values-y_pred.values)**2).sum(1) / n_cts)
    return rmse


def bootstrap_dists(corr_df, corr_gt_df, n_iter=1000, size=10):
    """
    Calculate the avg. distance to ground-truth (sub)-matrix based on random subsampling
    """
    if size == None:
        size = corr_df.shape[0]
    n = min(size, size)
    labels = corr_df.columns
    dists = np.zeros(n_iter)

    for i in range(n_iter):
        lbl = np.random.choice(corr_df.columns, n)
        A = corr_df.loc[lbl, lbl].values
        A_gt = corr_gt_df.loc[lbl, lbl].values
        dists[i] = _dist2gt(A, A_gt)

    return dists


def disp_rmse(y_true, y_preds, labels, title=None, return_rmse=False):
    """
    Boxplot of per-spot RMSEs for each prediction
    """
    n_spots, n_cts = y_true.shape
    rmse = np.array([
        np.sqrt(((y_true.values-y_pred.values)**2).sum(1) / n_cts)
        for y_pred in y_preds
    ])

    lbls = np.repeat(labels, n_spots)
    df = pd.DataFrame({
        'RMSE': rmses.flatten(),
        'Method': lbls
    })
    plt.figure(figsize=(10, 6))
    g = sns.boxplot(x='Method', y='RMSE', data=df)
    g.set_xticklabels(labels, rotation=60)
    plt.suptitle(title)
    plt.show()

    return rmses if return_rmse else None


def disp_corr(
    y_true, y_pred,
    outdir=None,
    figsize=(3.2, 3.2),
    fontsize=5,
    title=None,
    filename=None,
    savefig=False,
    format='png',
    return_corr=False
):
    """
    Calculate & plot correlation of cell proportion (or absolute cell abundance)
    between ground-truth & predictions (both [S x F])
    """

    assert y_true.shape[0] == y_pred.shape[0], 'Inconsistent sample sizes between ground-truth & prediction'
    if savefig:
        assert format == 'png' or format == 'eps' or format == 'svg', "Invalid saving format"

    v1 = y_true.values
    v2 = y_pred.values

    n_factor1, n_factor2 = v1.shape[1], v2.shape[1]
    corr = np.zeros((n_factor1, n_factor2))
    gt_corr = y_true.corr().values

    for i in range(n_factor1):
        for j in range(n_factor2):
            corr[i, j], _ = np.round(pearsonr(v1[:, i], v2[:, j]), 3)
 
    fig, ax = plt.subplots(1, 1, figsize=figsize, dpi=300)
    ax = sns.heatmap(
        corr, annot=True,
        cmap='RdBu_r', vmin=-1, vmax=1,
        annot_kws={"fontsize": fontsize},
        cbar_kws={'label': 'Cell type proportion corr.'},
        ax=ax
    )

    ax.set_xticks(np.arange(n_factor2) + 0.5)
    ax.set_yticks(np.arange(n_factor1) + 0.5)
    ax.set_xticklabels(y_pred.columns, rotation=90)
    ax.set_yticklabels(y_true.columns, rotation=0)
    ax.set_xlabel('Estimated proportion')
    ax.set_ylabel('Ground truth proportion')

    if title is not None:
        # ax.set_title(title+'\n'+'Distance = %.3f' % (dist2identity(corr)))
        ax.set_title(title + '\n' + 'Distance = %.3f' % (_calc_rmse(y_true, y_pred).mean()))

    for item in (ax.get_xticklabels() + ax.get_yticklabels()):
        item.set_fontsize(12)
    if savefig and (outdir is not None and filename is not None):
        if not os.path.exists(outdir):
            os.makedirs(outdir)
        fig.savefig(os.path.join(outdir, filename + '.' + format), bbox_inches='tight', format=format)
    plt.show()

    return corr if return_corr else None


def disp_prop_scatter(
    y_true, y_pred,
    outdir=None,
    filename=None,
    savefig=False,
    format='png'
):
    """
    Scatter plot of spot-wise proportion between ground-truth & predictions
    """
    assert y_true.shape == y_pred.shape, 'Inconsistent dimension between ground-truth & prediction'
    if savefig:
        assert format == 'png' or format == 'eps' or format == 'svg', "Invalid saving format"

    n_factors = y_true.shape[1]
    y_true_vals = y_true.values
    y_pred_vals = y_pred.values
    ncols = int(np.ceil(n_factors / 2))

    fig, (ax1, ax2) = plt.subplots(2, ncols, figsize=(2 * ncols, 4.4), dpi=300)

    for i in range(n_factors):
        v1 = y_true_vals[:, i]
        v2 = y_pred_vals[:, i]
        r2 = r2_score(v1, v2)

        v_stacked = np.vstack([v1, v2])
        den = gaussian_kde(v_stacked)(v_stacked)

        ax = ax1[i] if i < ncols else ax2[i % ncols]
        ax.scatter(v1, v2, c=den, s=.2, cmap='turbo', vmax=den.max() / 3)

        ax.set_aspect('equal')
        ax.spines['right'].set_visible(False)
        ax.spines['top'].set_visible(False)
        ax.axis('equal')

        # Only show ticks on the left and bottom spines
        ax.yaxis.set_ticks_position('left')
        ax.xaxis.set_ticks_position('bottom')

        ax.set_title(y_pred.columns[i])
        ax.annotate(r"$R^2$ = {:.3f}".format(r2), (0, 1), fontsize=8)

        ax.set_xlim([-0.1, 1.1])
        ax.set_ylim([-0.1, 1.1])
        ax.set_xticks(np.arange(0, 1.1, 0.5))
        ax.set_yticks(np.arange(0, 1.1, 0.5))

        ax.set_xlabel('Ground truth proportions')
        ax.set_ylabel('Predicted proportions')

    plt.tight_layout()
    if savefig and (outdir is not None and filename is not None):
        if not os.path.exists(outdir):
            os.makedirs(outdir)
        fig.savefig(os.path.join(outdir, filename + '.' + format), bbox_inches='tight', format=format)

    plt.show()




import numpy as np
import pandas as pd
import os.path
import sys
import scanpy as sc
from scipy.io import mmread
import anndata
from typing import Optional, List, Union


def _clean_up(df: pd.DataFrame) -> pd.DataFrame:
    """
    Remove rows and columns with all zeros from a DataFrame.
    
    Parameters
    ----------
    df : pd.DataFrame
        Input DataFrame to clean.
        
    Returns
    -------
    pd.DataFrame
        Cleaned DataFrame with rows and columns containing all zeros removed.
    """
    df = df.loc[df.sum(axis=1) > 0, :]
    df = df.loc[:, df.sum(axis=0) > 0]
    return df


def from_csv(counts_csv_file: str, delimiter: str = ",") -> pd.DataFrame:
    """
    Read gene expression data from a CSV file.
    
    Parameters
    ----------
    counts_csv_file : str
        Path to the CSV file containing gene expression data.
    delimiter : str, optional
        Delimiter used in the CSV file. Default is ','.
        
    Returns
    -------
    pd.DataFrame
        Gene expression data with rows as cells and columns as genes.
        Cells and genes with zero counts are removed.
    """
    # Read in csv file
    df = pd.read_csv(counts_csv_file, sep=delimiter, index_col=0)
    clean_df = _clean_up(df)
    return clean_df


def from_mtx(mtx_file: str, gene_name_file: str) -> pd.DataFrame:
    """
    Read gene expression data from a Matrix Market format file.
    
    Parameters
    ----------
    mtx_file : str
        Path to the Matrix Market file containing gene expression data.
    gene_name_file : str
        Path to the file containing gene names, one per line.
        
    Returns
    -------
    pd.DataFrame
        Gene expression data with rows as cells and columns as genes.
        Cells and genes with zero counts are removed.
    """
    # Read in mtx file
    count_matrix = mmread(mtx_file)

    gene_names = np.loadtxt(gene_name_file, dtype=np.dtype("S"))
    gene_names = np.array([gene.decode("utf-8") for gene in gene_names])

    # Convert to dense format
    df = pd.DataFrame(count_matrix.todense(), columns=gene_names)

    return _clean_up(df)


def from_10x(data_dir: Optional[str], use_ensemble_id: bool = True) -> pd.DataFrame:
    """
    Load data from 10X Genomics format.
    
    Parameters
    ----------
    data_dir : Optional[str]
        Directory containing the 10X Genomics output files:
        matrix.mtx, genes.tsv, and barcodes.tsv.
        If None, the current directory is used.
    use_ensemble_id : bool, optional
        If True, use Ensembl IDs as gene identifiers. 
        If False, use gene symbols. Default is True.
        
    Returns
    -------
    pd.DataFrame
        Gene expression data with rows as cells and columns as genes.
        Cells and genes with zero counts are removed.
    """
    # loads 10x sparse format data
    if data_dir is None:
        data_dir = "./"
    elif data_dir[len(data_dir) - 1] != "/":
        data_dir = data_dir + "/"

    filename_dataMatrix = os.path.expanduser(data_dir + "matrix.mtx")
    filename_genes = os.path.expanduser(data_dir + "genes.tsv")
    filename_cells = os.path.expanduser(data_dir + "barcodes.tsv")

    # Read in gene expression matrix (sparse matrix)
    # Rows = genes, columns = cells
    dataMatrix = mmread(filename_dataMatrix)

    # Read in row names (gene names / IDs)
    gene_names = np.loadtxt(filename_genes, delimiter="\t", dtype=bytes).astype(str)
    if use_ensemble_id:
        gene_names = [gene[0] for gene in gene_names]
    else:
        gene_names = [gene[1] for gene in gene_names]
    cell_names = np.loadtxt(filename_cells, delimiter="\t", dtype=bytes).astype(str)

    dataMatrix = pd.DataFrame(dataMatrix.todense(), columns=cell_names, index=gene_names)

    # combine duplicate genes
    if not use_ensemble_id:
        dataMatrix = dataMatrix.groupby(dataMatrix.index).sum()
    dataMatrix = dataMatrix.transpose()

    return _clean_up(dataMatrix)


def from_10x_HDF5(filename: str, genome: Optional[str] = None) -> pd.DataFrame:
    """
    Load data from 10X Genomics HDF5 format.
    
    Parameters
    ----------
    filename : str
        Path to the HDF5 file containing 10X Genomics data.
    genome : Optional[str], optional
        Name of the genome to load. If None, the first genome is used.
        
    Returns
    -------
    pd.DataFrame
        Gene expression data with rows as cells and columns as genes.
        Cells and genes with zero counts are removed.
    """
    ad = sc.read_10x_h5(filename, genome=genome, gex_only=True)

    dataMatrix = pd.DataFrame(ad.X.todense(), columns=ad.var_names, index=ad.obs_names)

    return _clean_up(dataMatrix)


def from_fcs(
    cls,
    fcs_file: str,
    cofactor: float = 5,
    metadata_channels: List[str] = [
        "Time",
        "Event_length",
        "DNA1",
        "DNA2",
        "Cisplatin",
        "beadDist",
        "bead1",
    ],
) -> pd.DataFrame:
    """
    Load data from Flow Cytometry Standard (FCS) format.
    
    Parameters
    ----------
    cls : object
        Class instance (unused, kept for compatibility).
    fcs_file : str
        Path to the FCS file to load.
    cofactor : float, optional
        Cofactor for arcsinh transformation. Default is 5.
    metadata_channels : List[str], optional
        List of metadata channel names to exclude from the returned data.
        
    Returns
    -------
    pd.DataFrame
        Processed cytometry data with metadata channels removed and
        optionally transformed using arcsinh.
        
    Notes
    -----
    This function requires the fcsparser package to be installed.
    If not installed, it will raise an ImportError with instructions.
    """
    try:
        import fcsparser
    except ImportError:
        raise ImportError(
            "The fcsparser package is required for reading FCS files. "
            "Please install it with: pip install fcsparser"
        )
    # Parse the fcs file
    text, data = fcsparser.parse(fcs_file)
    # Use view instead of newbyteorder for NumPy 2.0 compatibility
    data = data.astype(np.float64, copy=False)

    # Metadata and data
    metadata_channels = data.columns.intersection(metadata_channels)
    data_channels = data.columns.difference(metadata_channels)
    # metadata = data[metadata_channels]
    data = data[data_channels]

    # Transform if necessary
    if cofactor is not None or cofactor > 0:
        data = np.arcsinh(np.divide(data, cofactor))

    return data


r"""
Spatial transcriptomics analysis utilities.

This module provides comprehensive tools for analyzing spatial transcriptomics data
including spatial clustering, integration, trajectory analysis, and visualization.
Supported platforms include 10x Visium, Slide-seq, MERFISH, seqFISH, and others.

Spatial clustering and segmentation:
    pySTAGATE: Graph attention networks for spatial domains
    clusters: Spatial domain identification
    merge_cluster: Merge spatial clusters
    CAST: Cellular spatial organization analysis
    
Spatial integration and alignment:
    pySTAligner: Multi-sample spatial alignment
    Cal_Spatial_Net: Spatial neighborhood networks
    pySpaceFlow: Spatial flow analysis
    
Spatially variable genes:
    svg: Identify spatially variable genes
    
Spatial trajectory analysis:
    STT: Spatial transition tensor analysis
    CellMap: Cell mapping in spatial context
    CellLoc: Cell localization analysis
    
Spatial deconvolution:
    Tangram: Map single-cell data to spatial coordinates
    GASTON: Spatial deconvolution and analysis
    
Utility functions:
    Various spatial analysis tools and helper functions

Examples:
    >>> import omicverse as ov
    >>> import scanpy as sc
    >>> 
    >>> # Load spatial data
    >>> adata = sc.read_visium('path/to/visium/data')
    >>> 
    >>> # Identify spatial domains
    >>> ov.space.pySTAGATE(adata, n_domains=7, radius=50)
    >>> 
    >>> # Find spatially variable genes
    >>> adata = ov.space.svg(adata, mode='prost', n_svgs=2000)
    >>> 
    >>> # Spatial trajectory analysis
    >>> stt = ov.space.STT(adata, spatial_loc='spatial')
    >>> stt.stage_estimate()
    >>> stt.train(n_states=10)
    >>> 
    >>> # Spatial integration
    >>> aligner = ov.space.pySTAligner(adata_list)
    >>> aligner.spatial_alignment()
"""

from ._cluster import pySTAGATE,clusters,merge_cluster
from ._tangram import Tangram
from ._integrate import pySTAligner,Cal_Spatial_Net
from ._spaceflow import pySpaceFlow
from ._spatrio import CellMap,CellLoc
from ._stt import STT
from ._svg import svg
from ._cast import CAST
from ._tools import *
from ._gaston import GASTON
from ._commot import create_communication_anndata,update_classification_from_database


__all__ = [
    # Spatial clustering and domains
    'pySTAGATE',
    'clusters',
    'merge_cluster',
    'CAST',
    
    # Spatial integration and mapping
    'Tangram',
    'pySTAligner',
    'Cal_Spatial_Net',
    'pySpaceFlow',
    
    # Spatially variable genes
    'svg',
    
    # Spatial trajectory and dynamics
    'STT',
    'CellMap',
    'CellLoc',
    
    # Spatial deconvolution
    'GASTON',
    
    # Utility functions (imported from _tools)
    'create_communication_anndata',
    'update_classification_from_database',
]



import numpy as np
import scipy.sparse as sp


def load_dataset(file_name):
    """Load a graph from a Numpy binary file.

    Parameters
    ----------
    file_name : str
        Name of the file to load.

    Returns
    -------
    graph : dict
        Dictionary that contains:
            * 'A' : The adjacency matrix in sparse matrix format
            * 'X' : The attribute matrix in sparse matrix format
            * 'Z' : The community labels in sparse matrix format
            * Further dictionaries mapping node, class and attribute IDs

    """
    if not file_name.endswith('.npz'):
        file_name += '.npz'
    with np.load(file_name, allow_pickle=True) as loader:
        loader = dict(loader)
        A = sp.csr_matrix((loader['adj_matrix.data'], loader['adj_matrix.indices'],
                           loader['adj_matrix.indptr']), shape=loader['adj_matrix.shape'])

        if 'attr_matrix.data' in loader.keys():
            X = sp.csr_matrix((loader['attr_matrix.data'], loader['attr_matrix.indices'],
                               loader['attr_matrix.indptr']), shape=loader['attr_matrix.shape'])
        else:
            X = None

        Z = sp.csr_matrix((loader['labels.data'], loader['labels.indices'],
                           loader['labels.indptr']), shape=loader['labels.shape'])

        # Remove self-loops
        A = A.tolil()
        A.setdiag(0)
        A = A.tocsr()

        # Convert label matrix to numpy
        if sp.issparse(Z):
            Z = Z.toarray().astype(np.float32)

        graph = {
            'A': A,
            'X': X,
            'Z': Z
        }

        node_names = loader.get('node_names')
        if node_names is not None:
            node_names = node_names.tolist()
            graph['node_names'] = node_names

        attr_names = loader.get('attr_names')
        if attr_names is not None:
            attr_names = attr_names.tolist()
            graph['attr_names'] = attr_names

        class_names = loader.get('class_names')
        if class_names is not None:
            class_names = class_names.tolist()
            graph['class_names'] = class_names

        return graph



r"""
Bulk omics analysis utilities.

This module provides comprehensive tools for bulk RNA-seq analysis including:
- Differential expression analysis with DESeq2
- Gene co-expression network analysis (WGCNA)
- Pathway enrichment analysis (GSEA)
- Protein-protein interaction networks
- Batch effect correction
- Multi-omics integration

Classes:
    pyDEG: Differential expression analysis using DESeq2
    pyWGCNA: Weighted Gene Co-expression Network Analysis
    pyGSEA: Gene Set Enrichment Analysis
    pyPPI: Protein-Protein Interaction analysis
    pyTCGA: TCGA data analysis utilities

Functions:
    geneset_enrichment: Perform pathway enrichment analysis
    string_interaction: Retrieve STRING database interactions
    batch_correction: Correct for batch effects
    get_chm13_gene: Map genes to CHM13 reference

Examples:
    >>> import omicverse as ov
    >>> # Differential expression analysis
    >>> dds = ov.bulk.pyDEG(count_matrix, design_matrix)
    >>> dds.drop_duplicates_index()
    >>> dds.deseq2()
    >>> deg_results = dds.get_results()
    >>> 
    >>> # WGCNA analysis
    >>> wgcna = ov.bulk.pyWGCNA(adata)
    >>> wgcna.calculate_correlation()
    >>> wgcna.module_detection()
    """

# Heavy functionality lives in submodules and is imported lazily.
from ._Gene_module import pyWGCNA,readWGCNA
from ._Enrichment import pyGSEA,pyGSE,geneset_enrichment,geneset_plot,geneset_enrichment_GSEA,geneset_plot_multi,enrichment_multi_concat
from ._network import pyPPI,string_interaction,string_map,generate_G
from ._chm13 import get_chm13_gene,find_chm13_gene
from ._Deseq2 import pyDEG,deseq2_normalize,estimateSizeFactors,estimateDispersions,Matrix_ID_mapping,data_drop_duplicates_index
from ._tcga import pyTCGA
from ._combat import batch_correction

__all__ = [
    # Gene co-expression analysis
    'pyWGCNA',
    'readWGCNA',
    
    # Pathway enrichment analysis
    'pyGSEA',
    'pyGSE',
    'geneset_enrichment',
    'geneset_plot',
    'geneset_enrichment_GSEA',
    'geneset_plot_multi',
    'enrichment_multi_concat',
    
    # Protein-protein interaction networks
    'pyPPI',
    'string_interaction',
    'string_map',
    'generate_G',
    
    # Genome reference utilities
    'get_chm13_gene',
    'find_chm13_gene',
    
    # Differential expression analysis
    'pyDEG',
    'deseq2_normalize',
    'estimateSizeFactors',
    'estimateDispersions',
    'Matrix_ID_mapping',
    'data_drop_duplicates_index',
    
    # TCGA analysis
    'pyTCGA',
    
    # Batch correction
    'batch_correction',
]



r"""
NoCD: Non-overlapping Community Detection for single-cell data.

This module implements non-overlapping community detection algorithms specifically
designed for single-cell omics data. It provides neural network-based approaches
for identifying discrete cell populations and communities in high-dimensional
biological data.

Key features:
- Neural network-based community detection
- Non-overlapping cluster identification
- Specialized metrics for biological data evaluation
- Scalable training procedures for large datasets
- Integration with single-cell analysis workflows

Main components:
    data: Data loading and preprocessing utilities
    nn: Neural network architectures for community detection
    metrics: Evaluation metrics for clustering quality
    sampler: Data sampling strategies for training
    train: Training procedures and optimization
    utils: Utility functions and helper methods

Applications:
- Cell type identification in single-cell RNA-seq
- Population structure analysis
- Discrete state detection in developmental trajectories
- Batch effect-robust clustering

Examples:
    >>> import omicverse as ov
    >>> # Load and preprocess data
    >>> loader = ov.nocd.data.load_data(adata)
    >>> 
    >>> # Initialize neural network model
    >>> model = ov.nocd.nn.CommunityDetectionNet(
    ...     input_dim=adata.n_vars,
    ...     hidden_dim=64,
    ...     n_communities=10
    ... )
    >>> 
    >>> # Train the model
    >>> trainer = ov.nocd.train.Trainer(model, loader)
    >>> trainer.fit()
    >>> 
    >>> # Evaluate results
    >>> metrics = ov.nocd.metrics.evaluate_clustering(
    ...     predictions, ground_truth
    ... )

Notes:
    - Designed for discrete, non-overlapping community structure
    - Optimized for high-dimensional biological data
    - Provides both supervised and unsupervised training modes
    - Supports GPU acceleration for large-scale analysis
"""

from . import data
from . import nn
from . import metrics
from . import sampler
from . import train
from . import utils
